from utils_test import *
import uaibot as ub
import time
import sys
import matplotlib.pyplot as plt
import matplotlib
matplotlib.use('Qt5Agg')  
import os

print("Running Object to Object Smooth Distance Test")


projball1 = ub.Ball(radius=0.02,color='blue')
projball2 = ub.Ball(radius=0.02,color='red')

sim = ub.Simulation.create_sim_grid([projball1, projball2])



sum_time = 0
max_time = 0
cross_error_max = 0
N=100
eps=0.08, 
h=0.1
dmin0=0.2
dmin=0.05
eta=0.5
dt=0.01
Tmax=20
Nt = round(Tmax/dt)
No = 10
counter=0


hist_dist = []
hist_dist_0 = []
t = 0

for k in range(No):
    
    cont=True
    while cont:
        obj1 = generate_rand_object(color='red')
        obj2 = generate_rand_object(color='blue')
        type1 = ub.Utils.get_uaibot_type(obj1)
        type2 = ub.Utils.get_uaibot_type(obj2)
        
        if type1 == 'uaibot.PointCloud' and type2 == 'uaibot.PointCloud':
            cont=True
        else:
            cont = ub.Utils.compute_dist(obj1, obj2)[2]<dmin0

    sim.add(obj1)
    sim.add(obj2)
    
    if type1 != 'uaibot.PointCloud':
        htm1 = np.matrix(obj1.htm)
        obj1.add_ani_frame(0,htm=ub.Utils.trn([1000,1000,1000]))
        obj1.add_ani_frame(t,htm=htm1)
    else:
        obj1.add_ani_frame(0,0,0)
        obj1.add_ani_frame(t,0,np.shape(obj1.points)[1])
    
    if type2 != 'uaibot.PointCloud':
        htm2 = np.matrix(obj2.htm)
        obj2.add_ani_frame(0,htm=ub.Utils.trn([1000,1000,1000]))
        obj2.add_ani_frame(t,htm=htm2)    
    else:
        obj2.add_ani_frame(0,0,0)
        obj2.add_ani_frame(t,0,np.shape(obj2.points)[1])    

    for i in range(Nt):
        t += dt
        p1, p2, d, _ = ub.Utils.compute_dist(obj1, obj2, no_iter_max=2000, eps=0.02, h=0.1, tol=1e-5)
        p1_0, p2_0, d0, _ = ub.Utils.compute_dist(obj1, obj2)
        
        projball1.add_ani_frame(t,ub.Utils.trn(p1))
        projball2.add_ani_frame(t,ub.Utils.trn(p2))  
        
        
        w1 = np.matrix([np.cos(t), np.sin(t), np.cos(2*t)]).transpose()
        w2 = np.matrix([np.sin(2*t), np.sin(t), np.cos(0.5*t)]).transpose()
        
        alpha = max(d0-dmin,0)
        
        v1 = (1-alpha)*eta*(p1_0-p2_0)+alpha*np.matrix([np.sin(t), np.cos(t), 0]).transpose()
        v2 = (1-alpha)*eta*(p2_0-p1_0)+alpha*np.matrix([np.cos(t), np.cos(t), 0]).transpose()
        
        if type1 != 'uaibot.PointCloud':
            htm1_new = np.matrix(obj1.htm)
            htm1_new[0:3,-1] += v1*dt
            htm1_new = htm1_new * ub.Utils.rot(w1,np.linalg.norm(w1)*dt)
            obj1.add_ani_frame(t,htm=htm1_new) 
        
        if type2 != 'uaibot.PointCloud':
            htm2_new = np.matrix(obj2.htm)
            htm2_new[0:3,-1] += v2*dt   
            htm2_new = htm2_new * ub.Utils.rot(w2,np.linalg.norm(w2)*dt)
            obj2.add_ani_frame(t,htm=htm2_new) 

        sys.stdout.write("\rCompleted: "+str(round(100*counter/(Nt*No)))+"%")
        sys.stdout.flush()      
        
              
        hist_dist.append(d)
        hist_dist_0.append(d0)
        counter+=1



    if type1 != 'uaibot.PointCloud':
        obj1.add_ani_frame(t,htm=ub.Utils.trn([1000,1000,1000]))
    else:
        obj1.add_ani_frame(t,0,0)
            
    if type2 != 'uaibot.PointCloud':
        obj2.add_ani_frame(t,htm=ub.Utils.trn([1000,1000,1000]))
    else:
        obj2.add_ani_frame(t,0,0)

plt.plot(hist_dist)
plt.plot(hist_dist_0)
plt.show()


print("Test was a success!")   
    
    

current_folder = os.path.dirname(os.path.abspath(__file__))           
sim.save(current_folder,"test_o2o_smooth")



def _detach_object(self, obj):
    ind = -1
    for i in range(len(self.attached_objects)):
        if self.attached_objects[i][0] == obj:
            ind = i

    if ind == -1:
        raise Exception("The object was not found.")

    del self.attached_objects[ind]



from utils import *
import numpy as np
import os

from uaibot.graphics.meshmaterial import *
from uaibot.simobjects.pointcloud import *

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from uaibot.simobjects.pointcloud import PointCloud
    from uaibot.graphics.meshmaterial import MeshMaterial
    
from uaibot.utils.types import HTMatrix, Matrix, Vector, MetricObject
from typing import Optional, Tuple, List

class Box:
    """
  A box object.

  Parameters
  ----------
  htm : 4x4 numpy matrix
      The object's configuration
      (default: the same as the current HTM).

  name : string
      The object's name.
      (default: '' (automatic)).

  width : positive float
      The object's width, in meters.
      (default: 1).    

  depth : positive float
      The object's depth, in meters.
      (default: 1).  

  height : positive float
      The object's height, in meters.
      (default: 1).  

  color : string
      The object's color, a HTML - compatible string.
      (default: "red").

  opacity : float between 0 and 1
      The opacity. 1 = fully opaque, and 0 = transparent.

  mesh_material: 'MeshMaterial' object
      The object mesh material. If set to 'None', the default is used.
      (default: None).
  """

    #######################################
    # Attributes
    #######################################

    @property
    def width(self) -> float:
        """The box width, in meters."""
        return self._width

    @property
    def height(self) -> float:
        """The box height, in meters."""
        return self._height

    @property
    def depth(self) -> float:
        """The box depth, in meters."""
        return self._depth

    @property
    def name(self) -> str:
        """The object name."""
        return self._name

    @property
    def htm(self) -> "HTMatrix":
        """Object pose. A 4x4 homogeneous transformation matrix written is scenario coordinates."""
        return np.matrix(self._htm)

    @property
    def color(self) -> str:
        """Color of the object"""
        return self.mesh_material.color

    @property
    def mesh_material(self) -> "MeshMaterial":
        """Mesh material properties of the object"""
        return self._mesh_material

    #######################################
    # Constructor
    #######################################

    def __init__(self, htm: "HTMatrix" =np.identity(4), name : str ="", 
                 width: float =1, height: float =1, depth: float =1, color: str="red", opacity: float =1, \
                 mesh_material: Optional["MeshMaterial"] = None) -> "Box":

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(width) or width < 0:
            raise Exception("The parameter 'width' should be a positive float.")

        if not Utils.is_a_number(height) or height < 0:
            raise Exception("The parameter 'height' should be a positive float.")

        if not Utils.is_a_number(depth) or depth < 0:
            raise Exception("The parameter 'depth' should be a positive float.")

        if name=="":
            name="var_box_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a HTML-compatible color.")

        if not ((mesh_material is None) or (Utils.get_uaibot_type(mesh_material) == "uaibot.MeshMaterial")):
            raise Exception(
                "The parameter 'mesh_material' should be either 'None' or a 'uaibot.MeshMaterial' object.")

        if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
            raise Exception("The parameter 'opacity' should be a float between 0 and 1")
            # end error handling

        self._width = width
        self._height = height
        self._depth = depth
        self._htm = np.matrix(htm)
        self._name = name
        self._frames = []
        self._max_time = 0

        if mesh_material is None:
            self._mesh_material = MeshMaterial(color=color, opacity=opacity)
        else:
            self._mesh_material = mesh_material

        # Set initial total configuration
        self.set_ani_frame(self._htm)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Box with name '" + self.name + "': \n\n"
        string += " Width (m): " + str(self.width) + "\n"
        string += " Depth (m): " + str(self.depth) + "\n"
        string += " Height (m): " + str(self.height) + "\n"
        string += " Color: " + str(self.color) + "\n"
        string += " HTM: \n" + str(self.htm) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time: float, htm: Optional["HTMatrix"] = None) -> None:
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.
    htm : 4x4 numpy matrix
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """
        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(time) or time < 0:
            raise Exception("The parameter 'time' should be a positive float.")
        # end error handling

        f = [time, np.around(htm[0,0],4).item(), np.around(htm[0,2],4).item(), np.around(-htm[0,1],4).item(), np.around(htm[0,3],4).item(),
             np.around(htm[1,0],4).item(), np.around(htm[1,2],4).item(), np.around(-htm[1,1],4).item(), np.around(htm[1,3],4).item(),
             np.around(htm[2,0],4).item(), np.around(htm[2,2],4).item(), np.around(-htm[2,1],4).item(), np.around(htm[2,3],4).item(),
             0, 0, 0, 1]

        self._htm = htm
        self._frames.append(f)
        self._max_time = max(self._max_time, time)

    def set_ani_frame(self, htm: Optional["HTMatrix"] = None) -> None:
        """
    Reset object's animation queue and add a single configuration to the 
    object's animation queue.

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """

        if htm is None:
            htm = self._htm

        # Error handling
        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        # end error handling

        self._frames = []
        self.add_ani_frame(0, htm)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        string = "\n"
        string += "//BEGIN DECLARATION OF THE BOX '" + self.name + "'\n\n"
        string += self.mesh_material.gen_code(self.name) + "\n"
        string += "const var_" + self.name + " = new Box(" + str(self.width) + "," + str(
            self.height) + "," + str(self.depth) + "," + str(self._frames) + ", material_" + self.name + ");\n"
        string += "sceneElements.push(var_" + self.name + ");\n"
        string += "//USER INPUT GOES HERE"
        return string

    def copy(self) -> "Box":
        """Return a deep copy of the object, without copying the animation frames."""
        return Box(self.htm, self.name + "_copy", self.width, self.height, self.depth, self.color)

    def aabb(self, mode: str ='auto') -> "Box":
        """
    Compute an AABB (axis-aligned bounding box), considering the current orientation of the object.

    Parameters
    ----------
    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto') 
            
    Returns
    -------
     aabb: the AABB as a uaibot.Box object
    """

        if (mode == 'c++') or (mode=='auto' and os.environ['CPP_SO_FOUND']=='1'):
            obj_cpp = Utils.obj_to_cpp(self) 
            
        if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
            raise Exception("c++ mode is set, but .so file was not loaded!")
        
        if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
            p1 = self.width * self.htm[:,0] + self.depth * self.htm[:,1] + self.height * self.htm[:,2]
            p2 = -self.width * self.htm[:, 0] + self.depth * self.htm[:, 1] + self.height * self.htm[:, 2]
            p3 = self.width * self.htm[:, 0] - self.depth * self.htm[:, 1] + self.height * self.htm[:, 2]
            p4 = self.width * self.htm[:, 0] + self.depth * self.htm[:, 1] - self.height * self.htm[:, 2]

            w = np.max([abs(p1[0, 0]), abs(p2[0, 0]), abs(p3[0, 0]), abs(p4[0, 0])])
            d = np.max([abs(p1[1, 0]), abs(p2[1, 0]), abs(p3[1, 0]), abs(p4[1, 0])])
            h = np.max([abs(p1[2, 0]), abs(p2[2, 0]), abs(p3[2, 0]), abs(p4[2, 0])])
            
            return Box(name = "aabb_"+self.name, width= w, depth=d, height=h, htm=Utils.trn(self.htm[0:3,-1]),opacity=0.5)
        
        else:
            aabb = obj_cpp.get_aabb()
            return Box(name = "aabb_"+self.name, width= aabb.lx, depth=aabb.ly, height=aabb.lz, htm=Utils.trn(aabb.p),opacity=0.5) 


    def to_point_cloud(self, disc: float =0.025, mode: str = 'auto') -> "PointCloud":
        """
    Transform the object into a PointCloud object using the discretization 'delta'.

    Parameters
    ----------
    
    disc: positive float
        Discretization.
        (default: 0.025)

    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto') 
            
    Returns
    -------
     pointcloud: the pointcloud object.
    """

        if (mode == 'c++') or (mode=='auto' and os.environ['CPP_SO_FOUND']=='1'):
            obj_cpp = Utils.obj_to_cpp(self) 
            
        if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
            raise Exception("c++ mode is set, but .so file was not loaded!")
        
        if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
            P = np.matrix(np.zeros((3, 0)))

            T = round(2*np.pi*self.radius / disc)+1
            R = round(self.radius/disc)+1
            H = round(self.height / disc)+1


            for i in range(T):
                u = (2*np.pi)*i/(T-1)
                for j in range(H):
                    v = j/(H-1)

                    x = self.radius*np.cos(u)
                    y = self.radius*np.sin(u)
                    z = (-self.height/2 + v*self.height)
                    P = np.block([P, np.matrix([x,y,z]).transpose()])


            for i in range(R):
                v = self.radius * (i/(R-1))
                T = round(2 * np.pi * v / disc)
                for j in range(T):
                    u = (2*np.pi)*j/(T-1)

                    x = v * np.cos(u)
                    y = v * np.sin(u)
                    z = -self.height / 2
                    P = np.block([P, np.matrix([x, y, z]).transpose()])

                    x = v * np.cos(u)
                    y = v * np.sin(u)
                    z = self.height / 2
                    P = np.block([P, np.matrix([x, y, z]).transpose()])

            for i in range(np.shape(P)[1]):
                P[:,i] = self.htm[0:3,0:3]*P[:,i]+self.htm[0:3,-1]
                
            return PointCloud(points = P, color = self.color, size = disc/2)
        
        else:
            return PointCloud(points = obj_cpp.to_pointcloud(disc).points_gp, color = self.color, size = disc/2)
      
    # Compute distance to an object
    def compute_dist(self, obj: MetricObject,  p_init: Optional[Vector] = None, 
                     tol: float =0.001, no_iter_max: int =20, h: float =0, 
                     eps: float = 0, mode: str ='auto') -> Tuple[Vector, Vector, float, List]:
        """
    Compute Euclidean distance or differentiable distance between two objects.
    
    If h>0 or eps > 0, it computes the Euclidean distance and it uses GJK's algorithm.
    
    Else, it computes the differentiable distance through Generalized Alternating Projection (GAP).
    See the paper 'A Differentiable Distance Metric for Robotics Through Generalized Alternating Projection'.
    This only works in c++ mode, though.
    
    
    Parameters
    ----------
    obj : an object of type 'MetricObject' (see Utils.IS_METRIC)
        The other object for which we want to compute the distance.
        
    p_init : a 3D vector (3-element list/tuple, (3,1)/(1,3)/(3,)-shaped numpy matrix/numpy array) or None
        Initial point for closest point in this object. If 'None', is set to random.
        (default: None).
    
    tol : positive float
        Convergence criterion of GAP: it stops when ||a[k+1]-a[k]|| < tol.
        Only valid when h > 0 or eps > 0.
        (default: 0.001m).   

    no_iter_max : positive int 
        Maximum number of iterations of GAP.
        Only valid when h > 0 or eps > 0.
        (default: 20 iterations). 

    h : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0). 

    eps : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0). 

    mode : string
    'c++' for the c++ implementation, 'python' for the python implementation
    and 'auto' for automatic ('c++' is available, else 'python').
    (default: 'auto').
                                                    
    Returns
    -------
    point_this : 3 x 1 numpy matrix
        Closest point (Euclidean or differentiable) in this object.

    point_other : 3 x 1 numpy matrix
        Closest point (Euclidean or differentiable) in the other object.

    distance : float
        Euclidean or differentiable distance.
        
    hist_error: list of floats
        History of convergence error.    
                
    """
            
        return Utils.compute_dist(self, obj, p_init, tol, no_iter_max, h, eps, mode)

    
    # Compute the projection of a point into an object
    def projection(self, point: Vector, h: float =0, eps: float = 0, mode: str ='auto') -> Tuple[np.matrix, float]:
        """
    The projection of a point in the object, that is, the
    closest point in the object to a point 'point'.

    Parameters
    ----------
    point : a 3D vector (3-element list/tuple, (3,1)/(1,3)/(3,)-shaped numpy matrix/numpy array)
        The point for which the projection will be computed.

    h : positive float
        Smoothing parameter (only valid in c++ mode)
        (default: 0).            

    eps : positive float
        Smoothing parameter (only valid in c++ mode)
        (default: 0).      
        
    Returns
    -------
     proj_point : 3 x 1 numpy matrix
        The projection of the point 'point' in the object.

     d : positive float
        The distance between the object and 'point'.
    """


        if (mode == 'c++') or (mode=='auto' and os.environ['CPP_SO_FOUND']=='1'):
            obj_cpp = Utils.obj_to_cpp(self)
            
        if ( ( h > 0 or eps > 0) and ((mode == 'python') or ((mode=='auto' and os.environ['CPP_SO_FOUND']=='0')))):
            raise Exception("In Python mode, smoothing parameters 'h' and 'eps' must be set to 0!")
               
        if not Utils.is_a_number(h) or h < 0:
            raise Exception("The optional parameter 'h' must be a nonnegative number.")

        if not Utils.is_a_number(eps) or eps < 0:
            raise Exception("The optional parameter 'eps' must be a nonnegative number.")
        
        if not Utils.is_a_vector(point, 3):
            raise Exception("The parameter 'point' should be a 3D vector.")
        
        if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
            raise Exception("c++ mode is set, but .so file was not loaded!")

        # end error handling
        
        point_cvt = Utils.cvt(point)
        
        if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
            tpoint = self._htm[0:3, 0:3].T * (point_cvt - self._htm[0:3, 3])

            if abs(tpoint[0,0]) < self.width/2:
                x = tpoint[0,0]
                dx2 = 0
            else:
                x = self.width/2 if tpoint[0,0] > 0 else -self.width/2
                dx2 = (abs(tpoint[0,0]) - self.width/2)**2

            if abs(tpoint[1,0]) < self.depth/2:
                y = tpoint[1,0]
                dy2 = 0
            else:
                y = self.depth/2 if tpoint[1,0] > 0 else -self.depth/2
                dy2 = (abs(tpoint[1,0]) - self.depth/2)**2

            if abs(tpoint[2,0]) < self.height/2:
                z = tpoint[2,0]
                dz2 = 0
            else:
                z = self.height/2 if tpoint[2,0] > 0 else -self.height/2
                dz2 = (abs(tpoint[2,0]) - self.height/2)**2

            d = sqrt(dx2+dy2+dz2)

            return self._htm[0:3, 0:3] * np.matrix([[x], [y], [z]]) + self._htm[0:3, 3], d
        else:
            pr = obj_cpp.projection(Utils.cvt(point), h, eps)
            return Utils.cvt(pr.proj), pr.dist



from utils import *
import numpy as np
import os


from uaibot.simobjects.box import *

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from uaibot.simobjects.box import Box

    
from uaibot.utils.types import HTMatrix, Matrix, Vector, MetricObject
from typing import Optional, Tuple, List



if os.environ['CPP_SO_FOUND']=="1":
    import uaibot_cpp_bind as ub_cpp


class PointCloud:
    """
   A cloud of points to draw in the simulator.

   There is a fixed set of points ('points' attribute), and animation can be done by choosing, at each time,
   an interval [initial_ind, final_ind] that determines the range of points that will be displayed.

   Parameters
   ----------

   name : string
       The object's name.
       (default: '' (automatic)).

   size : positive float
       The size of each point in the point cloud.

   color : string
       A HTML-compatible color.

   points : a 3xm numpy array or matrix
       A matrix with 3 rows. The first row is the x coordinates, the second row the y coordinates, and the third row
       the z coordinates.

   """

    #######################################
    # Attributes
    #######################################

    @property
    def name(self):
        """The object name."""
        return self._name

    @property
    def size(self):
        """The size of each point in the point cloud."""
        return self._size

    @property
    def color(self):
        """Color of the object, a HTML-compatible string."""
        return self._color

    @property
    def points(self):
        """The points that compose the point cloud."""
        return np.array(self._points)

    @property
    def cpp_pointcloud(self):
        """The c++ version of the point cloud"""
        return self._cpp_pointcloud
    
    #######################################
    # Constructor
    #######################################

    @staticmethod
    def try_conversion(vector_list):
        if not isinstance(vector_list, list):
            raise Exception("Error")

        stacked_vectors = [] 

        for vec in vector_list:

            if not (isinstance(vec, np.ndarray) or isinstance(vec, list)):
                raise   Exception("Error")
            if isinstance(vec, list):
                if not ( len(vec) == 3):
                    raise   Exception("Error")
                else:
                    vec = np.matrix(vec).reshape((3,1))    
            if vec.shape == (3,):
                vec = vec.reshape(3, 1)  
            elif vec.shape == (1, 3):
                vec = vec.T 
            elif vec.shape == (3, 1):
                pass  
            else:
                raise  Exception("Error")

            stacked_vectors.append(vec)

        if not stacked_vectors:
            raise  Exception("Error")

        return np.matrix(np.hstack(stacked_vectors))



    def __init__(self, name="", points=[], size=0.1, color="blue"):

        # Error handling
        if not Utils.is_a_number(size) or size < 0:
            raise Exception("The parameter 'size' should be a positive float")


        if name=="":
            name="var_pointcloud_id_"+str(id(self))
            
        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a color")

        if isinstance(points, list) or (not Utils.is_a_matrix(points, 3)):
            try:
                self._points = PointCloud.try_conversion(points)
            except:
                raise Exception("The parameter 'points' should be a matrix with 3 rows or a list in which all the entries are 3D points (numpy or list of values).")
        else:
            self._points = np.matrix(points)
            
        # end error handling

        self._name = name
        self._size = size
        self._color = color
        self._frames = []
        self._max_time = 0
        
        if os.environ['CPP_SO_FOUND']=='1':
            list_of_points = [self._points[:, i:i+1] for i in range(self._points.shape[1])]
            self._cpp_pointcloud = ub_cpp.CPP_GeometricPrimitives.create_pointcloud(list_of_points)
        else:
            self._cpp_pointcloud = []

        self.add_ani_frame(0, 0, np.shape(self._points)[1])

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Point cloud '" + self._name + "' with " + str(np.shape(self.points)[1]) + " points: \n\n"
        string += " Size: " + str(self.size) + "\n"
        string += " Color: " + str(self._color) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time, initial_ind, final_ind):

        if (not Utils.is_a_number(time)) or time < 0:
            raise Exception("The parameter 'time' should be a nonnegative float.")

        if (not str(type(initial_ind)) == "<class 'int'>") or initial_ind < 0:
            raise Exception("The parameter 'initial_ind' should be a nonnegative integer.")

        if (not str(type(final_ind)) == "<class 'int'>") or final_ind < 0:
            raise Exception("The parameter 'final_ind' should be a nonnegative integer.")

        if initial_ind > final_ind:
            raise Exception("The parameter 'initial_ind' should be greater or equal than the parameter 'final_ind'.")

        if final_ind > len(self.points[0]):
            raise Exception("The parameter 'final_ind' should be at most " + str(len(self.points[0])) + ".")

        # Error handling

        # end error handling

        self._frames.append([time, initial_ind, final_ind])
        self._max_time = max(self._max_time, time)

    def set_ani_frame(self, time, initial_ind, final_ind):

        if (not Utils.is_a_number(time)) or time < 0:
            raise Exception("The parameter 'time' should be a nonnegative float.")

        if (not str(type(initial_ind)) == "<class 'int'>") or initial_ind < 0:
            raise Exception("The parameter 'initial_ind' should be a nonnegative integer.")

        if (not str(type(final_ind)) == "<class 'int'>") or final_ind < 0:
            raise Exception("The parameter 'final_ind' should be a nonnegative integer.")

        if initial_ind > final_ind:
            raise Exception("The parameter 'initial_ind' should be greater or equal than the parameter 'final_ind'.")

        if final_ind > len(self.points[0]):
            raise Exception("The parameter 'final_ind' should be at most " + str(len(self.points[0])) + ".")

        # Error handling

        # end error handling

        self._frames = []
        self.add_ani_frame(0, initial_ind, final_ind)
        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        string = "\n"
        string += "//BEGIN DECLARATION OF THE POINT CLOUD '" + self.name + "'\n\n"
        string += "const var_" + self.name + " = new PointCloud(" + str(np.around(np.matrix(self.points),4).tolist()) + ", " + str(
            self._frames) + ", '" + self.color + "', " + str(
            self.size) + ");\n"
        string += "sceneElements.push(var_" + self.name + ");\n"
        string += "//USER INPUT GOES HERE"

        return string

    def copy(self):
        return PointCloud(self.name + "_copy", np.matrix(self.points), self.size, self.color)

    def aabb(self, mode='auto'):
        """
    Compute an AABB (axis-aligned bounding box), considering the current orientation of the object.

    Parameters
    ----------
    mode : string
        'c++' for the c++ implementation, 'python' for the python implementation
        and 'auto' for automatic ('c++' is available, else 'python')
        (default: 'auto') 
            
    Returns
    -------
     aab: the AABB as a uaibot.Box object
    """

        if (mode == 'c++') or (mode=='auto' and os.environ['CPP_SO_FOUND']=='1'):
            obj_cpp = self._cpp_pointcloud
            
        if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
            raise Exception("c++ mode is set, but .so file was not loaded!")
        
        from simobjects.box import Box

        if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
            row_mins = self.points.min(axis=1) 
            row_maxs = self.points.max(axis=1)  
            xmin, ymin, zmin = row_mins
            xmax, ymax, zmax = row_maxs
            w = xmax-xmin
            d = ymax-ymin
            h = zmax-zmin
            center = np.matrix([(xmax+xmin)/2, (ymax+ymin)/2, (zmax+zmin)/2]).transpose()
            return Box(name = "aabb_"+self.name, width= w, depth=d, height=h, htm=Utils.trn(center),opacity=0.5)
        else:
            aabb = obj_cpp.get_aabb()
            return Box(name = "aabb_"+self.name, width= aabb.lx, depth=aabb.ly, height=aabb.lz, htm=Utils.trn(aabb.p),opacity=0.5)


    # Compute distance to an object
    def compute_dist(self, obj,  p_init=None, tol=0.001, no_iter_max=20, h=0, eps = 0, mode='auto'):
        """
    Compute Euclidean distance or differentiable distance between two objects.
    
    If h>0 or eps > 0, it computes the Euclidean distance and it uses GJK's algorithm.
    
    Else, it computes the differentiable distance through Generalized Alternating Projection (GAP).
    See the paper 'A Differentiable Distance Metric for Robotics Through Generalized Alternating Projection'.
    This only works in c++ mode, though.
    
    
    Parameters
    ----------
    obj : an object of type 'MetricObject' (see Utils.IS_METRIC)
        The other object for which we want to compute the distance.
        
    p_init : a 3D vector (3-element list/tuple, (3,1)/(1,3)/(3,)-shaped numpy matrix/numpy array) or None
        Initial point for closest point in this object. If 'None', is set to random.
        (default: None).
    
    tol : positive float
        Convergence criterion of GAP: it stops when ||a[k+1]-a[k]|| < tol.
        Only valid when h > 0 or eps > 0.
        (default: 0.001m).   

    no_iter_max : positive int 
        Maximum number of iterations of GAP.
        Only valid when h > 0 or eps > 0.
        (default: 20 iterations). 

    h : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0). 

    eps : nonnegative float
        h parameter in the generalized distance function.
        If h=0 and eps=0, it is simply the Euclidean distance.
        (default: 0). 

    mode : string
    'c++' for the c++ implementation, 'python' for the python implementation
    and 'auto' for automatic ('c++' is available, else 'python').
    (default: 'auto').
                                                    
    Returns
    -------
    point_this : 3 x 1 numpy matrix
        Closest point (Euclidean or differentiable) in this object.

    point_other : 3 x 1 numpy matrix
        Closest point (Euclidean or differentiable) in the other object.

    distance : float
        Euclidean or differentiable distance.
        
    hist_error: list of floats
        History of convergence error.    
                
    """        
        
        return Utils.compute_dist(self, obj, p_init, tol, no_iter_max, h, eps, mode)
    
    # Compute the projection of a point into an object
    def projection(self, point, h=0, eps = 0, mode='auto') -> Tuple[np.matrix, float]:
        """
    The projection of a point in the object, that is, the
    closest point in the object to a point 'point'.

    Parameters
    ----------
    point : a 3D vector (3-element list/tuple, (3,1)/(1,3)/(3,)-shaped numpy matrix/numpy array)
        The point for which the projection will be computed.

    h : positive float
        Smoothing parameter (only valid in c++ mode)
        (default: 0).            

    eps : positive float
        Smoothing parameter (only valid in c++ mode)
        (default: 0).      
        
    Returns
    -------
     proj_point : 3 x 1 numpy matrix
        The projection of the point 'point' in the object.

     d : positive float
        The distance between the object and 'point'.
    """


        if ( ( h > 0 or eps > 0) and ((mode == 'python') or ((mode=='auto' and os.environ['CPP_SO_FOUND']=='0')))):
            raise Exception("In Python mode, smoothing parameters 'h' and 'eps' must be set to 0!")
               
        if not Utils.is_a_number(h) or h < 0:
            raise Exception("The optional parameter 'h' must be a nonnegative number.")

        if not Utils.is_a_number(eps) or eps < 0:
            raise Exception("The optional parameter 'eps' must be a nonnegative number.")
        
        if not Utils.is_a_vector(point, 3):
            raise Exception("The parameter 'point' should be a 3D vector.")
        
        if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
            raise Exception("c++ mode is set, but .so file was not loaded!")

        # end error handling
        if mode == 'python' or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
            raise Exception("Projection to point cloud available only for c++ mode!")
        else:
            pr = self._cpp_pointcloud.projection(Utils.cvt(point), h, eps)
            return Utils.cvt(pr.proj), pr.dist



def _gen_code(self):
    # Inject code
    string = "\n"
    string += "//BEGIN DECLARATION OF THE ROBOT '" + self.name + "'\n\n"

    string += "const object3d_base_" + self.name + "_list = [];\n\n"

    if not (self.list_object_3d_base is None):
        for i in range(len(self.list_object_3d_base)):
            string += self.list_object_3d_base[i].gen_code(self.name + "_obj_base_" + str(i))
            string += "object3d_base_" + self.name + "_list.push(object3d_" + self.name + "_obj_base_" + str(i) + ");\n"

        string += "\n"

    string += "const list_links_" + self.name + " = [];\n\n"

    for i in range(len(self.links)):
        string += self.links[i].gen_code(self.name)
        string += "list_links_" + self.name + ".push(link_" + str(i) + "_" + self.name + ");\n\n"

    string += "const htm_" + self.name + "_base_0 = new Matrix4(); \n"
    string += "htm_" + self.name + "_base_0.set(" \
              + str(self.htm_base_0[0,0]) + "," + str(self.htm_base_0[0,1]) + "," + str(
        self.htm_base_0[0,2]) + "," + str(
        self.htm_base_0[0,3]) + "," \
              + str(self.htm_base_0[1,0]) + "," + str(self.htm_base_0[1,1]) + "," + str(
        self.htm_base_0[1,2]) + "," + str(
        self.htm_base_0[1,3]) + "," \
              + str(self.htm_base_0[2,0]) + "," + str(self.htm_base_0[2,1]) + "," + str(
        self.htm_base_0[2,2]) + "," + str(
        self.htm_base_0[2,3]) + "," \
              + str(self.htm_base_0[3,0]) + "," + str(self.htm_base_0[3,1]) + "," + str(
        self.htm_base_0[3,2]) + "," + str(
        self.htm_base_0[3,3]) + ");\n"

    string += "const htm_" + self.name + "_n_eef = new Matrix4(); \n"
    string += "htm_" + self.name + "_n_eef.set(" \
              + str(self.htm_n_eef[0,0]) + "," + str(self.htm_n_eef[0,1]) + "," + str(
        self.htm_n_eef[0,2]) + "," + str(
        self.htm_n_eef[0,3]) + "," \
              + str(self.htm_base_0[1,0]) + "," + str(self.htm_n_eef[1,1]) + "," + str(
        self.htm_n_eef[1,2]) + "," + str(
        self.htm_n_eef[1,3]) + "," \
              + str(self.htm_n_eef[2,0]) + "," + str(self.htm_n_eef[2,1]) + "," + str(
        self.htm_n_eef[2,2]) + "," + str(
        self.htm_n_eef[2,3]) + "," \
              + str(self.htm_n_eef[3,0]) + "," + str(self.htm_n_eef[3,1]) + "," + str(
        self.htm_n_eef[3,2]) + "," + str(
        self.htm_n_eef[3,3]) + ");\n"

    string += "const var_" + self.name + " = new Robot(object3d_base_" + self.name + "_list, list_links_" + self.name + "," + str(
        self._frames) + ", htm_" + self.name + "_base_0, htm_" + self.name + "_n_eef,"+ ( "true" if self.eef_frame_visible else "false" )+");\n"
    string += "sceneElements.push(var_" + self.name + ");\n"
    string += "//USER INPUT GOES HERE"

    return string



from utils import *
import numpy as np
import os

# Function used for task function/task Jacobian
def _task_function(self, htm_tg, q=None, htm=None, mode='auto'):

    if q is None:
        q = self.q

    if htm is None:
        htm = self.htm

    # Error handling
    if mode not in ['python','c++','auto']:
        raise Exception("The parameter 'mode' should be 'python,'c++', or 'auto'.")
       
    if not Utils.is_a_matrix(htm_tg, 4, 4):
        raise Exception("The parameter 'htm_tg' should be a 4x4 homogeneous transformation matrix.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

    n = len(self.links)
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    # end error handling
    if mode=='c++' and os.environ['CPP_SO_FOUND']=='0':
        raise Exception("c++ mode is set, but .so file was not loaded!")
    # end error handling

    if mode == 'python'  or (mode=='auto' and os.environ['CPP_SO_FOUND']=='0'):
        return _task_function_python(self, htm_tg, Utils.cvt(q), htm)
    else:
        task_res = self.cpp_robot.fk_task(Utils.cvt(q), htm, htm_tg)
        return  Utils.cvt(task_res.task), Utils.cvt(task_res.jac_task)

             
            
def _task_function_python(self, htm_tg, q=None, htm=None):


    n = len(self.links)


    p_des = htm_tg[0:3, 3]
    x_des = htm_tg[0:3, 0]
    y_des = htm_tg[0:3, 1]
    z_des = htm_tg[0:3, 2]

    jac_eef, htm_eef = self.jac_geo(q, "eef", htm, mode='python')
    p_eef = htm_eef[0:3, 3]
    x_eef = htm_eef[0:3, 0]
    y_eef = htm_eef[0:3, 1]
    z_eef = htm_eef[0:3, 2]

    r = np.matrix(np.zeros((6,1)))
    r[0:3,0] = p_eef - p_des
    r[3] = max(1 - x_des.T * x_eef, 0)
    r[4] = max(1 - y_des.T * y_eef, 0)
    r[5] = max(1 - z_des.T * z_eef, 0)

    n = len(self.links)
    jac_r = np.matrix(np.zeros((6, n)))
    jac_r[0:3, :] = jac_eef[0:3, :]
    jac_r[3, :] = x_des.T * Utils.S(x_eef) * jac_eef[3:6, :]
    jac_r[4, :] = y_des.T * Utils.S(y_eef) * jac_eef[3:6, :]
    jac_r[5, :] = z_des.T * Utils.S(z_eef) * jac_eef[3:6, :]



    return r, jac_r



import numpy as np
from utils import *
from uaibot.utils.types import HTMatrix, Matrix, Vector, MetricObject
from typing import Optional, Tuple, List

class PointLight:
    """
   A point light.

   Parameters
   ----------

   name : string
       The object's name.
       (default: '' (automatic)).

   color : string
       A HTML-compatible color.
       (default: 'white').

   intensity : positive float
       The light intensity.
       (default: 1).

   htm : 4x4 numpy matrix
      The object's configuration.
      (default: the same as the current HTM).

   max_distance : nonnegative float
       The maximum distance in which the light can act, in meters.
       If set to 0, this distance is infinite.
       (default: 0).

   """

    #######################################
    # Attributes
    #######################################

    @property
    def name(self) -> str:
        """The object name."""
        return self._name

    @property
    def color(self) -> str:
        """The light color."""
        return self._color

    @property
    def intensity(self) -> float:
        """The light intensity"""
        return self._intensity

    @property
    def htm(self) -> "HTMatrix":
        """Object pose. A 4x4 homogeneous transformation matrix written is scenario coordinates."""
        return np.matrix(self._htm)

    @property
    def max_distance(self) -> float:
        """The light maximum distance (meters)."""
        return self._max_distance

    #######################################
    # Constructor
    #######################################

    def __init__(self, htm: "HTMatrix" =np.identity(4), name: str ="", color: str ="white", 
                 intensity: float =1, max_distance: float =0) -> "PointLight":

        # Error handling

        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if name=="":
            name="var_pointlight_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a color.")

        if not Utils.is_a_number(intensity) or intensity < 0:
            raise Exception("The parameter 'intensity' should be a nonnegative float.")

        if not Utils.is_a_number(max_distance) or max_distance < 0:
            raise Exception("The parameter '_max_distance' should be a nonnegative float.")

        # end error handling

        self._name = name
        self._color = color
        self._intensity = intensity
        self._htm = htm
        self._max_distance = max_distance
        self._frames = []
        self._max_time = 0

        self.add_ani_frame(0, self.htm, self.color, self.intensity, self.max_distance)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Point light with name '" + self.name + "': \n\n"
        string += " Color: " + str(self.color) + "\n"
        string += " Position: " + str(self.htm[0:3,3].tolist()) + "\n"
        string += " Intensity: " + str(self.intensity) + "\n"
        string += " Maximum distance: " + str(self.max_distance) + "\n"

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time: float, htm: Optional[HTMatrix]=None, 
                      color: Optional[str] = None, intensity: Optional[float] = None, 
                      max_distance: Optional[float] = None) -> None:
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.

    htm : 4x4 numpy matrix
        The object's configuration
        (default: the same as the current HTM).

    color : string
       A HTML-compatible color.
       (default: the same as the current color).

    intensity : positive float
       The light intensity.
       (default: the same as the current intensity).

    max_distance : nonnegative float
       The maximum distance in which the light can act, in meters.
       If set to 0, this distance is infinite.
       (default: the same as the current max_distance).

    Returns
    -------
    None
    """
        if color is None:
            color = self.color
        if intensity is None:
            intensity = self.intensity
        if htm is None:
            htm = self._htm
        if max_distance is None:
            max_distance = self.max_distance

        # Error handling
        if not Utils.is_a_number(time) or time < 0:
            raise Exception("The parameter 'time' should be a positive float.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a color.")

        if not Utils.is_a_number(intensity) or intensity < 0:
            raise Exception("The parameter 'intensity' should be a nonnegative float.")

        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(max_distance) or max_distance < 0:
            raise Exception("The parameter '_max_distance' should be a nonnegative float.")
        # end error handling

        f = [time, color, intensity, max_distance, htm[0:3, 3].tolist()]

        self._color = color
        self._intensity = intensity
        self._htm = htm
        self._max_distance = max_distance
        self._frames.append(f)
        self._max_time = max(self._max_time, time)

    # Set config. Restart animation queue
    def set_ani_frame(self, htm: Optional[HTMatrix]=None, 
                      color: Optional[str] = None, intensity: Optional[float] = None, 
                      max_distance: Optional[float] = None) -> None:
        """
    Reset object's animation queue and add a single configuration to the
    object's animation queue.

    Parameters
    ----------

    htm : 4x4 numpy array
        The object's configuration
        (default: the same as the current HTM).

    color : string
       A HTML-compatible color.
       (default: the same as the current color).

    intensity : positive float
       The light intensity.
       (default: the same as the current intensity).

    max_distance : nonnegative float
       The maximum distance in which the light can act, in meters.
       If set to 0, this distance is infinite.
       (default: the same as the current max_distance).


    Returns
    -------
    None
    """

        if color is None:
            color = self.color
        if intensity is None:
            intensity = self.intensity
        if htm is None:
            htm = self._htm
        if max_distance is None:
            max_distance = self.max_distance

        # Error handling
        if not Utils.is_a_number(time) or time < 0:
            raise Exception("The parameter 'time' should be a positive float.")

        if not Utils.is_a_color(color):
            raise Exception("The parameter 'color' should be a color.")

        if not Utils.is_a_number(intensity) or intensity < 0:
            raise Exception("The parameter 'intensity' should be a nonnegative float.")

        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        if not Utils.is_a_number(max_distance) or max_distance < 0:
            raise Exception("The parameter '_max_distance' should be a nonnegative float.")
        # end error handling

        self._frames = []
        self.add_ani_frame(0, htm, color, intensity, max_distance)
        self._max_time = 0


    def gen_code(self):
        """Generate code for injection."""

        string = "\n"
        string += "//BEGIN DECLARATION OF THE POINT LIGHT '" + self.name + "'\n\n"
        string += "const var_" + self.name + " = new PointLightSim(" + str(self._frames) + ");\n"
        string += "sceneElements.push(var_" + self.name + ");\n"
        string += "//USER INPUT GOES HERE"
        return string



import numpy as np
from utils import *
from uaibot.utils.types import HTMatrix, Matrix, Vector, MetricObject, GroupableObject
from typing import Optional, Tuple, List

class Group:
    """
   A group of other 'simobject' objects.

   When the objects are grouped, it is assumed that the scenario frame is the parent frame of all objects.
   This means that when the group is created, it stores the htm of each object at that moment, htm_base_object[i].
   When the htm of the whole group is changed to 'htm', it changes the htm of each member i of the group to
   htm @ htm_base_object[i]. Thus, be sure that, when the group is created, all the objects are in the correct pose.

   To see which type of objects are groupable, see the constant 'Utils.IS_GROUPABLE'.

   Parameters
   ----------
   htm : 4x4 numpy matrix
       The object's configuration.
       (default: the same as the current HTM).

   name : string
       The object's name.
       (default: '' (automatic)).

   list_of_objects : list of objects that are in 'Utils.IS_GROUPABLE'
       List of objects. Each must be an object of the type contained in the constant list 'Utils.IS_GROUPABLE'.

   """

    #######################################
    # Attributes
    #######################################

    @property
    def name(self) -> str:
        """Name of the object."""
        return self._name

    @property
    def list_of_objects(self) -> List["GroupableObject"]:
        """The list of objects contained in the group."""
        return self._list_of_objects

    @property
    def htm(self) -> "HTMatrix":
        """Group pose. A 4x4 homogeneous transformation matrix written is scenario coordinates."""
        return self._htm

    @property
    def htm_base_object(self) -> List["HTMatrix"]:
        """The pose of each object, written in scenario coordinates, when the group was created."""
        return self._htm_base_object

    #######################################
    # Constructor
    #######################################

    def __init__(self, list_of_objects: List["GroupableObject"], name: str="", 
                 htm: "HTMatrix" =np.identity(4)) -> "Group":

        # Error handling

        if name=="":
            name="var_group_id_"+str(id(self))

        if not (Utils.is_a_name(name)):
            raise Exception(
                "The parameter 'name' should be a string. Only characters 'a-z', 'A-Z', '0-9' and '_' are allowed. It should not begin with a number.")

        if not ((str(type(list_of_objects)) == "<class 'list'>")):
            raise Exception(
                "The parameter 'list_of_objects' should be a list of simple objects (Utils.is_a_simple_object(obj)==True).")

        for obj in list_of_objects:
            if not (Utils.is_a_groupable_object(obj)):
                raise Exception(
                    "The parameter 'list_of_objects' should be a list of the following objects: " + str(
                        Utils.IS_GROUPABLE) + ".")

        if not Utils.is_a_matrix(htm, 4, 4):
            raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")

        for i in range(len(list_of_objects)):
            if list_of_objects[i].name == name:
                raise Exception("The name of the group ('" + name + "') is the same name of one of its elements.")

            for j in range(i + 1, len(list_of_objects)):
                if list_of_objects[i].name == list_of_objects[j].name:
                    raise Exception("The are two objects in the list (indexes " + str(i) + " and " + str(
                        j) + ") that have the same name.")

        # end error handling

        self._name = name
        self._list_of_objects = list_of_objects
        self._htm = htm
        self._max_time = 0

        self._htm_base_object = []
        for obj in self.list_of_objects:
            self._htm_base_object.append(obj.htm)

        # Set initial configuration
        self.set_ani_frame(self.htm)

    #######################################
    # Std. Print
    #######################################

    def __repr__(self):

        string = "Group '" + self.name + "': \n\n"
        string += "List of objects: "
        for obj in self.list_of_objects:
            string += "'" + obj.name + "', "

        return string

    #######################################
    # Methods
    #######################################

    def add_ani_frame(self, time: float, htm: Optional["HTMatrix"] = None) -> None:
        """
    Add a single configuration to the object's animation queue.

    Parameters
    ----------
    time: positive float
        The timestamp of the animation frame, in seconds.
    htm : 4x4 numpy matrix
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """

        for i in range(len(self.list_of_objects)):
            self.list_of_objects[i].add_ani_frame(time=time, htm=htm @ self.htm_base_object[i])
            self._max_time = max(self._max_time, self.list_of_objects[i]._max_time)

    def set_ani_frame(self, htm: Optional["HTMatrix"] = None) -> None:
        """
    Reset object's animation queue and add a single configuration to the
    object's animation queue.

    Parameters
    ----------
    htm : 4x4 numpy matrix
        The object's configuration
        (default: the same as the current HTM).

    Returns
    -------
    None
    """

        for i in range(len(self.list_of_objects)):
            self.list_of_objects[i].set_ani_frame(htm=htm @ self.htm_base_object[i])

        self._max_time = 0

    def gen_code(self):
        """Generate code for injection."""

        self._max_time = 0
        for i in range(len(self.list_of_objects)):
            self._max_time = max(self._max_time, self.list_of_objects[i]._max_time)

        string = "\n"
        string += "//BEGIN DECLARATION OF THE GROUP '" + self.name + "'\n\n"
        for obj in self.list_of_objects:
            string += obj.gen_code() + "\n"

        string = string.replace("//USER INPUT GOES HERE", "")
        string += "\n//USER INPUT GOES HERE"

        return string



import numpy as np
import uaibot as ub
import numpy as np
from scipy.optimize import linprog
from scipy.spatial import HalfspaceIntersection, ConvexHull

def is_bounded(A, b):
    n_dim = A.shape[1]

    for i in range(n_dim):
        c = np.zeros(n_dim)
        c[i] = 1  

        res = linprog(c, A_ub=A, b_ub=b, method='highs', bounds=[(None, None)] * n_dim)

        if res.status == 3 or abs(res.x[i])>1e3: 
            return False  

        c[i] = -1  

        res = linprog(c, A_ub=A, b_ub=b, method='highs', bounds=[(None, None)] * n_dim)

        if res.status == 3 or abs(res.x[i])>1e3: 
            return False  
        
    return True  

def generate_bounded_polytope(n_halfspaces=10, size = 1.0):
    while True:
        A = np.random.uniform(-1, 1, (n_halfspaces, 3))  # Each row is a normal vector
        A /= np.linalg.norm(A, axis=1, keepdims=True)
        b = size*np.abs(np.random.uniform(0.1, 0.3, (n_halfspaces,)))
        
        if is_bounded(A, b):
            return A, b

def generate_rand_object(objtype = -1, size = 1.0, spread = 1.0, color='red'):
    
    if objtype == -1:
        coin = np.random.randint(0,5)
    else:
        coin = objtype

    htm = ub.Utils.htm_rand([-spread*1.5,-spread*1.5,spread*0.5],[spread*1.5,spread*1.5,spread*1.5])
    
    if coin==0:
        r = size* np.random.uniform(0.3,1.0)
        obj = ub.Ball(htm=htm, radius=r, color=color)
        
        return obj
        
    if coin==1:
        w = size*np.random.uniform(0.3,0.8)
        d = size*np.random.uniform(0.3,0.8)
        h = size*np.random.uniform(0.3,0.8)
        obj = ub.Box(htm=htm, width=w, depth=d, height=h, color=color)
        
        return obj


    if coin==2:
        r = size*np.random.uniform(0.3,0.8)
        h = size*np.random.uniform(0.3,0.8)
        obj = ub.Cylinder(htm=htm, radius = r, height=h, color=color)
        
        return obj

        
    if coin==3:
        n = np.random.randint(500,1000)  
        m1 = np.matrix(np.random.uniform(low=0, high=1, size=(3, n))) 
        m2 = np.matrix(np.random.uniform(low=0, high=1, size=(3, n)))
        w = size*np.random.uniform(0.3,0.8)
        d = size*np.random.uniform(0.3,0.8)
        h = size*np.random.uniform(1.0,1.5)
        pc = np.matrix([w,d,h]).transpose()
        
        w = size*np.random.uniform(0.3,0.8)
        d = size*np.random.uniform(0.3,0.8)
        h = size*np.random.uniform(1.0,1.5)
        
        points = np.diag([w,d,h])*htm[0:3,0:3]*(m1+m2)+pc
        obj = ub.PointCloud(points = points, color=color)
        
        return obj
 
        
    if coin==4:
        A, b = generate_bounded_polytope(size = size)
        obj = ub.ConvexPolytope(A = A, b = b, color=color)
        obj.add_ani_frame(0,htm)
        
        return obj
            
        



from utils import *
import numpy as np


# Analytic Jacobian
def _jac_ana(self, q=None, htm=None):
    if q is None:
        q = self.q
    if htm is None:
        htm = np.matrix(self.htm)

    n = len(self.links)

    # Error handling
    if not Utils.is_a_vector(q, n):
        raise Exception("The parameter 'q' should be a " + str(n) + " dimensional vector.")

    if not Utils.is_a_matrix(htm, 4, 4):
        raise Exception("The parameter 'htm' should be a 4x4 homogeneous transformation matrix.")
    # end error handling

    jac_geo, htm_eef = self.jac_geo(Utils.cvt(q),'eef',htm)

    alpha, beta, gamma = Utils.euler_angles(htm_eef)

    sa = np.sin(alpha)
    ca = np.cos(alpha)
    cb = np.cos(beta)
    tb = np.tan(beta)

    transf = np.matrix([[-ca*tb, -sa*tb, 1],
                        [-sa   ,     ca, 0],
                        [ ca/cb,  sa/cb, 0]])

    jac_ana = np.matrix(np.zeros((6,n)))
    jac_ana[0:3,:] = jac_geo[0:3,:]
    jac_ana[3:6,:] = transf * jac_geo[3:6,:]

    phi = np.matrix([[alpha],[beta],[gamma]])

    return jac_ana, htm_eef, phi



__version__ = '0.0.21'

import os
import sys

# Get the directory of the current __init__.py file
current_dir = os.path.dirname(__file__)

# Add the directory of __init__.py to sys.path
if current_dir not in sys.path:
    sys.path.insert(0, current_dir)



try:
    import uaibot_cpp_bind as ub_cpp
    os.environ['CPP_SO_FOUND'] = '1'
except ImportError:
    print("INFO: CPP .so not found! Only python mode is available!")
    os.environ['CPP_SO_FOUND'] = '0'

from .robot import *
from .demo import *
from .utils import *
from .simulation import *
from .graphics import *
from .simobjects import *
from sfm import *

try:
    from .utils import Utils
    from .robot import Robot
    from .simulation import Simulation
    from .simobjects import Box, Ball, ConvexPolytope, Cylinder, PointCloud
    from .graphics import MeshMaterial, Texture
    from .demo import Demo
except ImportError:
    pass

__all__ = [
    "Utils", "Robot", "Simulation", "Demo",
    "Box", "Ball", "ConvexPolytope", "Cylinder", "PointCloud",
    "MeshMaterial", "Texture",
]

import numpy as np
np.set_printoptions(
    precision=4,   
    linewidth=80,  
    suppress=True  
)







from utils import Utils

from graphics.meshmaterial import MeshMaterial
from graphics.model3d import Model3D
from simobjects.box import Box
from simobjects.cylinder import Cylinder
from robot.links import Link

import numpy as np


def _create_davinci_arm1(color, opacity, name='davinci_arm1'):

    if not Utils.is_a_color(color):
        raise Exception(
            "The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception(
            "The parameter 'opacity' should be a float between 0 and 1.")

    theta2 = np.deg2rad(1.875)
    theta3 = np.deg2rad(3.25 - 1.875)
    theta4 = np.deg2rad(75.25)
    theta5 = np.deg2rad(28.75 + 90)
    theta6 = np.deg2rad(-45)
    theta7 = np.deg2rad(94 + 180)
    theta8 = np.deg2rad(-170.5)
    theta9 = np.deg2rad(61.5)

    d3 = -96e-3
    d5 = -96e-3 * 1.9
    d6 = (431.8 * 1.417)/1000
    d9 = -3.75e-2
    d10 = 3e-2

    alpha5 = np.pi*(1/2 + 1/9)
    alpha6 = np.pi/2
    alpha9 = np.pi/2
    alpha10 = np.pi/2*0

    a2 = 0.3
    a3 = 0.415
    a4 = 0.415
    a6 = 3.1e-3
    a7 = 0.27
    a8 = 0.474
    a9 = 0.098

    # Passive joints 1, 2, 3, 4, 5, 8, 9
    # Active joints 6, 7, 10
    link_info = np.array([
        # "theta" rotation in z
        [0, 0,   0,  0,      0,      0,  0,  0,      0,   0],
        # "d" translation in z
        [0, 0,  d3,  0,     d5,     d6,  0,  0,     d9, d10],
        # "alfa" rotation in x
        [0, 0,   0,  0, alpha5, alpha6,  0,  0, alpha9,   0],
        # "a" translation in x
        [0, a2, a3, a4,      0,     a6, a7, a8,     a9,   0],
        # joint type
        [1, 0,   0,  0,      0,      0,  0,  0,      0,   1]
    ])

    scale = 1
    n = link_info.shape[1]
    base_3d_obj = []
    link_3d_obj = []
    mesh = MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[
                        0.5, 0.5], color=color, opacity=opacity, side="DoubleSide")
    b1 = 0.1
    b2 = 0.45
    b3 = 1.1

    Q00 = Utils.trn([b1, b3, -b2]) * Utils.rotx(-np.pi/2) * \
        Utils.rotz(np.pi)  # change reference frame
    Q01 = Q00 * Utils.rotz(link_info[0, 0]) * Utils.trn([0, 0, link_info[1, 0]]) * Utils.rotx(link_info[2, 0]) * Utils.trn(
        [link_info[3, 0], 0, 0])
    Q02 = Q01 * (Utils.rotz(link_info[0, 0] + theta2) * Utils.trn([0, 0, link_info[1, 1]]) * Utils.rotx(link_info[2, 1]) * Utils.trn(
        [link_info[3, 1], 0, 0]))
    Q03 = Q02 * (Utils.rotz(link_info[0, 1] + theta3) * Utils.trn([0, 0, link_info[1, 2]]) * Utils.rotx(link_info[2, 2]) * Utils.trn(
        [link_info[3, 2], 0, 0]))
    Q04 = Q03 * (Utils.rotz(link_info[0, 2] + theta4) * Utils.trn([0, 0, link_info[1, 3]]) * Utils.rotx(link_info[2, 3]) * Utils.trn(
        [link_info[3, 3], 0, 0]))
    Q05 = Q04 * (Utils.rotz(link_info[0, 3] + theta5) * Utils.trn([0, 0, link_info[1, 4]]) * Utils.rotx(link_info[2, 4]) * Utils.trn(
        [link_info[3, 4], 0, 0]))
    Q06 = Q05 * (Utils.rotz(link_info[0, 4] + theta6) * Utils.trn([0, 0, link_info[1, 5]]) * Utils.rotx(link_info[2, 5]) * Utils.trn(
        [link_info[3, 5], 0, 0]))
    Q07 = Q06 * (Utils.rotz(link_info[0, 5] + theta7) * Utils.trn([0, 0, link_info[1, 6]]) * Utils.rotx(link_info[2, 6]) * Utils.trn(
        [link_info[3, 6], 0, 0]))
    Q08 = Q07 * (Utils.rotz(link_info[0, 6] + theta8) * Utils.trn([0, 0, link_info[1, 7]]) * Utils.rotx(link_info[2, 7]) * Utils.trn(
        [link_info[3, 7], 0, 0]))
    Q09 = Q08 * (Utils.rotz(link_info[0, 7] + theta9) * Utils.trn([0, 0, link_info[1, 8]]) * Utils.rotx(link_info[2, 8]) * Utils.trn(
        [link_info[3, 8], 0, 0]))
    Q010 = Q09 * (Utils.rotz(link_info[0, 8]) * Utils.trn([0, 0, link_info[1, 9]]) * Utils.rotx(link_info[2, 9]) * Utils.trn(
        [link_info[3, 9], 0, 0]))

    link1_mth = Utils.inv_htm(Q01)
    link_3d_obj.append([
        # torre movel
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/4.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
        # cilindro conector
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/5.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
    ])

    link2_mth = Utils.inv_htm(Q02)
    link_3d_obj.append([
        # ligacao de cilindros curta
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/8.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh),
    ])

    link3_mth = Utils.inv_htm(Q03)
    link_3d_obj.append([
        # cilindro conector
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/59.obj',
            scale=scale, htm=link3_mth, mesh_material=mesh),
        # ligacao de cilindros longa
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/9.obj',
            scale=scale, htm=link3_mth, mesh_material=mesh),
    ])

    link4_mth = Utils.inv_htm(Q04)
    link_3d_obj.append([
        # ligacao de cilindros longa
        Model3D(

            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/11.obj',
            scale=scale, htm=link4_mth, mesh_material=mesh),
    ])

    link5_mth = Utils.inv_htm(Q05)
    link_3d_obj.append([
        # cotovelo
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/10.obj',
            scale=scale, htm=link5_mth, mesh_material=mesh),
    ])

    link6_mth = Utils.inv_htm(Q06)
    link_3d_obj.append([
        # pá dobrada
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/38.obj',
            scale=scale, htm=link6_mth, mesh_material=mesh),
    ])

    link7_mth = Utils.inv_htm(Q07)
    link_3d_obj.append([
        # conecta pá com bumerangue
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/19.obj',
            scale=scale, htm=link7_mth, mesh_material=mesh),
    ])

    link8_mth = Utils.inv_htm(Q08)
    link_3d_obj.append([
        # bumerangue
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/21.obj',
            scale=scale, htm=link8_mth, mesh_material=mesh),

    ])

    link9_mth = Utils.inv_htm(Q09)
    link_3d_obj.append([
        # envoltoria agulha
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/43.obj',
            scale=scale, htm=link9_mth, mesh_material=mesh),
    ])

    link10_mth = Utils.inv_htm(Q010)
    link_3d_obj.append([
        # guia
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/45.obj',
                scale=scale, htm=link10_mth, mesh_material=mesh),
        # agulha
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/54.obj',
                scale=scale, htm=link10_mth, mesh_material=mesh),
        # abridor de garrafa
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/51.obj',
            scale=scale, htm=link10_mth, mesh_material=mesh),
        # botao
        Model3D(
            url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/52.obj',
                scale=scale, htm=link10_mth, mesh_material=mesh),
    ])

    col_model = [[], [], [], [], [], [], [], [], [], []]

    col_model[0].append(Box(htm=Utils.trn([0, 0.02, -0.33]),
                            name=name + "_C0_0", width=0.14, height=0.4, depth=0.2, color="red", opacity=0.3))
    col_model[0].append(Cylinder(htm=Utils.trn([0, 0, -0.14]) @ Utils.rotz(3.14 / 2) @ Utils.rotx(3.14),
                                 name=name + "_C0_1", radius=0.075, height=0.19, color="red", opacity=0.3))

    col_model[1].append(Cylinder(htm=Utils.trn([0, 0, 0.02]) @ Utils.rotz(np.pi/2) @ Utils.rotx(np.pi),
                                 name=name + "_C1_0", radius=0.075, height=0.12, color="black", opacity=0.3))
    col_model[1].append(Cylinder(htm=Utils.trn([-0.3, 0, 0.02]) @ Utils.rotz(np.pi/2) @ Utils.rotx(np.pi),
                                 name=name + "_C1_1", radius=0.075, height=0.12, color="black", opacity=0.3))
    col_model[1].append(Box(htm=Utils.trn([-0.15, 0, 0.02]) @ Utils.roty(np.pi/2),
                            name=name + "_C1_2", width=0.11, height=0.3, depth=0.12, color="black", opacity=0.3))

    col_model[2].append(Box(htm=Utils.trn([-0.200, 0, -0.12]) @ Utils.roty(np.pi/2),
                            name=name + "_C2_0", width=0.11, height=0.4, depth=0.12, color="green", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([0, 0, -0.12]) @ Utils.rotz(np.pi/2) @ Utils.rotx(np.pi),
                                 name=name + "_C2_1", radius=0.075, height=0.12, color="green", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([-0.415, 0, -0.065]) @ Utils.rotz(np.pi/2) @ Utils.rotx(np.pi),
                                 name=name + "_C2_2", radius=0.075, height=0.23, color="green", opacity=0.3))

    col_model[3].append(Box(htm=Utils.trn([-0.200, 0, 0]) @ Utils.roty(np.pi/2),
                            name=name + "_C3_0", width=0.1, height=0.4, depth=0.12, color="blue", opacity=0.3))
    col_model[3].append(Cylinder(htm=Utils.trn([0, 0, -0.005]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C3_1", radius=0.075, height=0.15, color="blue", opacity=0.3))
    col_model[3].append(Cylinder(htm=Utils.trn([-0.415, 0, 0.01]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C3_2", radius=0.075, height=0.13, color="blue", opacity=0.3))

    col_model[4].append(Cylinder(htm=Utils.trn([0, 0, 0]) @ Utils.rotx(-alpha5),
                                 name=name + "_C4_0", radius=0.075, height=0.2, color="magenta", opacity=0.3))
    col_model[4].append(Box(htm=Utils.trn([0, -0.02, 0.12]) @ Utils.roty(np.pi/2),
                            name=name + "_C4_1", width=0.2, height=0.14, depth=0.14, color="magenta", opacity=0.3))

    col_model[5].append(Box(htm=Utils.trn([0, -0.355, -0.05]) @ Utils.roty(np.pi/2),
                            name=name + "_C5_0", width=0.2, height=0.08, depth=0.05, color="orange", opacity=0.3))
    col_model[5].append(Box(htm=Utils.trn([0, -0.15, -0.125]) @ Utils.roty(np.pi/2),
                            name=name + "_C5_1", width=0.05, height=0.09, depth=0.4, color="orange", opacity=0.3))

    col_model[6].append(Box(htm=Utils.trn([-0.2, 0.025, -0.028]),
                            name=name + "_C6_0", width=0.25, height=0.14, depth=0.15, color="cyan", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.02, 0.025, -0.09]),
                            name=name + "_C6_1", width=0.14, height=0.02, depth=0.15, color="cyan", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.019, 0.023, 0.018]),
                            name=name + "_C6_2", width=0.14, height=0.02, depth=0.15, color="cyan", opacity=0.3))

    col_model[7].append(Box(htm=Utils.trn([-0.42, 0.08, -0.0365]) @ Utils.rotz(np.pi/3),
                            name=name + "_C7_0", width=0.23, height=0.082, depth=0.13, color="#88264a", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.22, 0.08, -0.0365]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C7_1", width=0.11, height=0.09, depth=0.39, color="#88264a", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.023, 0.01, -0.073]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C7_2", width=0.08, height=0.01, depth=0.04, color="#88264a", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.023, 0.01, 0.0]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C7_3", width=0.08, height=0.01, depth=0.04, color="#88264a", opacity=0.3))
    col_model[7].append(Cylinder(htm=Utils.trn([0, 0, 0.0]),
                                 name=name + "_C7_4", radius=0.035, height=0.01, color="#88264a", opacity=0.3))
    col_model[7].append(Cylinder(htm=Utils.trn([0, -0.001, -0.073]),
                                 name=name + "_C7_5", radius=0.035, height=0.01, color="#88264a", opacity=0.3))

    col_model[8].append(Box(htm=Utils.trn([-0.052, -0.001, -0.042]),
                            name=name + "_C8_0", width=0.02, height=0.31, depth=0.12, color="Brown", opacity=0.3))
    col_model[8].append(Cylinder(htm=Utils.trn([-0.1, 0, 0]) @ Utils.rotx(np.pi/2),
                                 name=name + "_C8_1", radius=0.035, height=0.056, color="Brown", opacity=0.3))
    col_model[8].append(Box(htm=Utils.trn([-0.08, 0, 0]),
                            name=name + "_C8_2", width=0.04, height=0.07, depth=0.056, color="Brown", opacity=0.3))
    col_model[8].append(Box(htm=Utils.trn([-0.03, 0, 0.075]),
                            name=name + "_C8_3", width=0.04, height=0.05, depth=0.04, color="Brown", opacity=0.3))
    col_model[8].append(Cylinder(htm=Utils.trn([0, 0, 0.075]),
                                 name=name + "_C8_4", radius=0.02, height=0.056, color="Brown", opacity=0.3))

    col_model[9].append(Box(htm=Utils.trn([-0.005, 0.0, -0.2]),
                            name=name + "_C9_0", width=0.07, height=0.4, depth=0.12, color="MidnightBlue", opacity=0.3))
    col_model[9].append(Cylinder(htm=Utils.trn([0, 0, 0.163]),
                                 name=name + "_C9_1", radius=0.01, height=0.4, color="MidnightBlue", opacity=0.3))

    links = []
    for i in range(n):
        links.append(Link(i, theta=link_info[0, i], d=link_info[1, i], alpha=link_info[2, i], a=link_info[3, i], joint_type=link_info[4, i],
                          list_model_3d=link_3d_obj[i]))
        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    #     1  2  3  4  5  6  7  8  9 10
    q0 = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
    htm_n_eef = Utils.trn([0, 0, 0.3025])
    htm_base_0 = Utils.trn([-b1, -b2, b3])

    # Create joint limits
    joint_limits = np.matrix([[-0.11, 0.636], [-np.pi, np.pi], [-np.pi, np.pi], [-np.deg2rad(162), np.deg2rad(162)], [-np.pi, np.pi],
                             [-np.pi, np.pi], [-np.pi, np.pi], [-np.deg2rad(175), np.deg2rad(50)], [-np.deg2rad(108), np.deg2rad(70)], [-0.08, 0.05]])

    return links, base_3d_obj, htm_base_0, htm_n_eef, q0, joint_limits


def _create_davinci_arm2(color, opacity, name='davinci_arm2'):

    if not Utils.is_a_color(color):
        raise Exception(
            "The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception(
            "The parameter 'opacity' should be a float between 0 and 1.")

    theta2 = np.deg2rad(-3)
    theta3 = np.deg2rad(98.5)
    theta4 = np.deg2rad(108.5)
    theta5 = np.deg2rad(-44)
    theta6 = np.deg2rad(-85.5)
    theta7 = np.deg2rad(-171)
    theta8 = np.deg2rad(61.7)

    d2 = 96e-3 * 1.2
    d4 = -96e-3 * 0.925
    d5 = 431.8e-3 * 1.419
    d8 = -2.9e-2

    alpha4 = np.deg2rad(90 + 36.5)
    alpha5 = np.pi/2
    alpha8 = np.deg2rad(90)

    a2 = 0.415
    a3 = 0.415
    a5 = -8e-4
    a6 = 0.27
    a7 = 0.474
    a8 = 0.098

    # Passive joints 1, 2, 3, 4, 7, 8
    # Active joints 5, 6, 9
    link_info = np.array([
        # "theta" rotation in z
        [0,  0,   0,       0,      0,      0,      0,       0, 0],
        # 0,  "d" translation in z
        [0, d2,   0,      d4,     d5,      0,      0,      d8, 0],
        # 0,  "alfa" rotation in x
        [0,  0,   0,  alpha4, alpha5,      0,      0,  alpha8, 0],
        # 0,  "a" translation in x
        [0, a2,  a3,       0,     a5,     a6,     a7,      a8, 0],
        # 0,  joint type
        [1,  0,   0,       0,      0,      0,      0,       0, 1]
    ])

    scale = 1
    n = link_info.shape[1]
    base_3d_obj = []
    link_3d_obj = []
    mesh = MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[
                        0.5, 0.5], color=color, opacity=opacity, side="DoubleSide")
    b1 = -0.086
    b2 = 0.2634
    b3 = 1.3

    Q00 = Utils.trn([b1, b3, -b2]) * Utils.rotx(-np.pi/2) * \
        Utils.rotz(np.pi)  # change reference frame
    Q01 = Q00 * (Utils.rotz(link_info[0, 0]) * Utils.trn([0, 0, link_info[1, 0]]) * Utils.rotx(link_info[2, 0]) * Utils.trn(
        [link_info[3, 0], 0, 0]))
    Q02 = Q01 * (Utils.rotz(link_info[0, 0] + theta2) * Utils.trn([0, 0, link_info[1, 1]]) * Utils.rotx(link_info[2, 1]) * Utils.trn(
        [link_info[3, 1], 0, 0]))
    Q03 = Q02 * (Utils.rotz(link_info[0, 1] + theta3) * Utils.trn([0, 0, link_info[1, 2]]) * Utils.rotx(link_info[2, 2]) * Utils.trn(
        [link_info[3, 2], 0, 0]))
    Q04 = Q03 * (Utils.rotz(link_info[0, 2] + theta4) * Utils.trn([0, 0, link_info[1, 3]]) * Utils.rotx(link_info[2, 3]) * Utils.trn(
        [link_info[3, 3], 0, 0]))
    Q05 = Q04 * (Utils.rotz(link_info[0, 3] + theta5) * Utils.trn([0, 0, link_info[1, 4]]) * Utils.rotx(link_info[2, 4]) * Utils.trn(
        [link_info[3, 4], 0, 0]))
    Q06 = Q05 * (Utils.rotz(link_info[0, 4] + theta6) * Utils.trn([0, 0, link_info[1, 5]]) * Utils.rotx(link_info[2, 5]) * Utils.trn(
        [link_info[3, 5], 0, 0]))
    Q07 = Q06 * (Utils.rotz(link_info[0, 5] + theta7) * Utils.trn([0, 0, link_info[1, 6]]) * Utils.rotx(link_info[2, 6]) * Utils.trn(
        [link_info[3, 6], 0, 0]))
    Q08 = Q07 * (Utils.rotz(link_info[0, 6] + theta8) * Utils.trn([0, 0, link_info[1, 7]]) * Utils.rotx(link_info[2, 7]) * Utils.trn(
        [link_info[3, 7], 0, 0]))
    Q09 = Q08 * (Utils.rotz(link_info[0, 7]) * Utils.trn([0, 0, link_info[1, 8]]) * Utils.rotx(link_info[2, 8]) * Utils.trn(
        [link_info[3, 8], 0, 0]))

    link1_mth = Utils.inv_htm(Q01)
    link_3d_obj.append([
        # torre movel
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/58.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
        # cilindro conector
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/63.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
    ])

    link2_mth = Utils.inv_htm(Q02)
    link_3d_obj.append([
        # ligacao de cilindros longa
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/64.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh),
    ])

    link3_mth = Utils.inv_htm(Q03)
    link_3d_obj.append([
        # ligacao de cilindros longa
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/61.obj',
                scale=scale, htm=link3_mth, mesh_material=mesh),
    ])

    link4_mth = Utils.inv_htm(Q04)
    link_3d_obj.append([
        # cotovelo
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/14.obj',
                scale=scale, htm=link4_mth, mesh_material=mesh),
    ])

    link5_mth = Utils.inv_htm(Q05)
    link_3d_obj.append([
        # pá dobrada
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/36.obj',
                scale=scale, htm=link5_mth, mesh_material=mesh),
    ])

    link6_mth = Utils.inv_htm(Q06)
    link_3d_obj.append([
        # conecta pá com bumerangue
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/34.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh),
    ])

    link7_mth = Utils.inv_htm(Q07)
    link_3d_obj.append([
        # bumerangue
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/39.obj',
                scale=scale, htm=link7_mth, mesh_material=mesh),
    ])

    link8_mth = Utils.inv_htm(Q08)
    link_3d_obj.append([
        # envoltoria agulha
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/42.obj',
                scale=scale, htm=link8_mth, mesh_material=mesh),
    ])

    link9_mth = Utils.inv_htm(Q09)
    link_3d_obj.append([
        # guia
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/23.obj',
                    scale=scale, htm=link9_mth, mesh_material=mesh),
        # agulha
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/55.obj',
                scale=scale, htm=link9_mth, mesh_material=mesh),
        # abridor de garrafa
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/25.obj',
                scale=scale, htm=link9_mth, mesh_material=mesh),
        # botao
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/53.obj',
                scale=scale, htm=link9_mth, mesh_material=mesh),
    ])

    col_model = [[], [], [], [], [], [], [], [], []]

    col_model[0].append(Box(htm=Utils.trn([-0.02, 0.006*0, -0.43]) @ Utils.rotz(np.pi/2),
                            name=name + "_C0_0", width=0.14, height=0.4, depth=0.2, color="red", opacity=0.3))
    col_model[0].append(Cylinder(htm=Utils.trn([0, 0, -0.16]) @ Utils.rotz(3.14 / 2) @ Utils.rotx(3.14),
                                 name=name + "_C0_1", radius=0.075, height=0.19, color="red", opacity=0.3))

    col_model[1].append(Box(htm=Utils.trn([-0.200, 0, -0.11]) @ Utils.roty(np.pi/2),
                            name=name + "_C1_0", width=0.1, height=0.4, depth=0.12, color="green", opacity=0.3))
    col_model[1].append(Cylinder(htm=Utils.trn([0, 0, -0.12]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C1_1", radius=0.075, height=0.12, color="green", opacity=0.3))
    col_model[1].append(Cylinder(htm=Utils.trn([-0.415, 0, -0.11]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C1_2", radius=0.075, height=0.13, color="green", opacity=0.3))

    col_model[2].append(Box(htm=Utils.trn([-0.200, 0, 0.005]) @ Utils.roty(np.pi/2),
                            name=name + "_C2_0", width=0.1, height=0.4, depth=0.12, color="blue", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([0, 0, 0.01]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C2_1", radius=0.075, height=0.12, color="blue", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([-0.415, 0, 0.01]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C2_2", radius=0.075, height=0.13, color="blue", opacity=0.3))

    col_model[3].append(Cylinder(htm=Utils.trn([0, -0.025, 0.01]) @ Utils.rotx(-alpha4),
                                 name=name + "_C3_0", radius=0.075, height=0.12, color="magenta", opacity=0.3))
    col_model[3].append(Box(htm=Utils.trn([0, -0.02, 0.12]) @ Utils.roty(np.pi/2),
                            name=name + "_C3_1", width=0.2, height=0.14, depth=0.14, color="magenta", opacity=0.3))

    col_model[4].append(Box(htm=Utils.trn([0, -0.355, -0.05]) @ Utils.roty(np.pi/2),
                            name=name + "_C4_0", width=0.2, height=0.08, depth=0.05, color="orange", opacity=0.3))
    col_model[4].append(Box(htm=Utils.trn([0, -0.15, -0.125]) @ Utils.roty(np.pi/2),
                            name=name + "_C4_1", width=0.05, height=0.09, depth=0.4, color="orange", opacity=0.3))

    col_model[5].append(Box(htm=Utils.trn([-0.2, 0.025, -0.025]),
                            name=name + "_C5_0", width=0.25, height=0.14, depth=0.15, color="cyan", opacity=0.3))
    col_model[5].append(Box(htm=Utils.trn([-0.02, 0.025, -0.088]),
                            name=name + "_C5_1", width=0.14, height=0.02, depth=0.15, color="cyan", opacity=0.3))
    col_model[5].append(Box(htm=Utils.trn([-0.019, 0.025, 0.019]),
                            name=name + "_C5_2", width=0.14, height=0.02, depth=0.15, color="cyan", opacity=0.3))

    col_model[6].append(Box(htm=Utils.trn([-0.42, 0.08, -0.034]) @ Utils.rotz(np.pi/3),
                            name=name + "_C6_0", width=0.23, height=0.082, depth=0.13, color="#88264a", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.22, 0.08, -0.034]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C6_1", width=0.11, height=0.09, depth=0.39, color="#88264a", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.023, 0.01, 0.005]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C6_2", width=0.08, height=0.01, depth=0.03, color="#88264a", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.023, 0.01, -0.068]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C6_3", width=0.08, height=0.01, depth=0.03, color="#88264a", opacity=0.3))
    col_model[6].append(Cylinder(htm=Utils.trn([0, 0, -0.068]),
                                 name=name + "_C6_4", radius=0.035, height=0.01, color="#88264a", opacity=0.3))
    col_model[6].append(Cylinder(htm=Utils.trn([0, 0, 0.005]),
                                 name=name + "_C6_5", radius=0.035, height=0.01, color="#88264a", opacity=0.3))

    col_model[7].append(Box(htm=Utils.trn([-0.052, 0.0, -0.042]),
                            name=name + "_C7_0", width=0.02, height=0.31, depth=0.12, color="Brown", opacity=0.3))
    col_model[7].append(Cylinder(htm=Utils.trn([-0.1, -0.005, -0.001]) @ Utils.rotx(np.pi/2),
                                 name=name + "_C7_1", radius=0.035, height=0.056, color="Brown", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.08, -0.005, 0]),
                            name=name + "_C7_2", width=0.04, height=0.07, depth=0.056, color="Brown", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.03, -0.001, 0.073]),
                            name=name + "_C7_3", width=0.04, height=0.05, depth=0.04, color="Brown", opacity=0.3))
    col_model[7].append(Cylinder(htm=Utils.trn([0, -0.001, 0.074]),
                                 name=name + "_C7_4", radius=0.02, height=0.056, color="Brown", opacity=0.3))

    col_model[8].append(Box(htm=Utils.trn([-0.005, 0.0, -0.22]),
                            name=name + "_C8_0", width=0.07, height=0.4, depth=0.12, color="MidnightBlue", opacity=0.3))
    col_model[8].append(Cylinder(htm=Utils.trn([0, 0, 0.135]),
                                 name=name + "_C8_1", radius=0.01, height=0.4, color="MidnightBlue", opacity=0.3))

    links = []
    for i in range(n):
        links.append(Link(i, theta=link_info[0, i], d=link_info[1, i], alpha=link_info[2, i], a=link_info[3, i], joint_type=link_info[4, i],
                          list_model_3d=link_3d_obj[i]))
        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    #     1  2  3  4  5  6  7  8  9
    q0 = [0, 0, 0, 0, 0, 0, 0, 0, 0]
    htm_n_eef = Utils.trn([0, 0, 0.3323])  # np.identity(4)
    htm_base_0 = Utils.trn([-b1, -b2, b3])  # np.identity(4)

    # Create joint limits
    joint_limits = np.matrix([[-0.1, 0.53], [-4*np.pi, 4*np.pi], [-np.deg2rad(162), np.deg2rad(162)], [-4*np.pi, 4*np.pi], [-4*np.pi, 4*np.pi],
                             [-4*np.pi, 4*np.pi], [-np.deg2rad(175), np.deg2rad(50)], [-np.deg2rad(108), np.deg2rad(70)], [-0.05, 0.079]])

    return links, base_3d_obj, htm_base_0, htm_n_eef, q0, joint_limits


def _create_davinci_arm3(color, opacity, name='davinci_arm3'):

    if not Utils.is_a_color(color):
        raise Exception(
            "The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception(
            "The parameter 'opacity' should be a float between 0 and 1.")

    theta2 = np.deg2rad(-3)
    theta3 = np.deg2rad(98.5)
    theta4 = np.deg2rad(108.5)
    theta5 = np.deg2rad(-44)
    theta6 = np.deg2rad(-85.5)
    theta7 = np.deg2rad(-171)
    theta8 = np.deg2rad(61.7)

    d2 = 96e-3 * 1.2
    d4 = -96e-3 * 0.925
    d5 = 431.8e-3 * 1.419
    d8 = -2.9e-2

    alpha4 = np.deg2rad(90 + 36.5)
    alpha5 = np.pi/2
    alpha8 = np.deg2rad(90)

    a2 = 0.415
    a3 = 0.415
    a5 = -8e-4
    a6 = 0.27
    a7 = 0.474
    a8 = 0.098

    # Passive joints 1, 2, 3, 4, 7, 8
    # Active joints 5, 6, 9
    link_info = np.array([
        # "theta" rotation in z
        [0,  0,   0,       0,      0,      0,      0,       0, 0],
        # 0,  "d" translation in z
        [0, d2,   0,      d4,     d5,      0,      0,      d8, 0],
        # 0,  "alfa" rotation in x
        [0,  0,   0,  alpha4, alpha5,      0,      0,  alpha8, 0],
        # 0,  "a" translation in x
        [0, a2,  a3,       0,     a5,     a6,     a7,      a8, 0],
        # 0,  joint type
        [1,  0,   0,       0,      0,      0,      0,       0, 1]
    ])

    scale = 1
    n = link_info.shape[1]
    base_3d_obj = []
    link_3d_obj = []
    mesh = MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[
                        0.5, 0.5], color=color, opacity=opacity, side="DoubleSide")
    b1 = -0.086
    b2 = 0.2634
    b3 = 1.3

    Q00 = Utils.trn([b1, b3, -b2]) * Utils.rotx(-np.pi/2) * \
        Utils.rotz(np.pi)  # change reference frame
    Q01 = Q00 * (Utils.rotz(link_info[0, 0]) * Utils.trn([0, 0, link_info[1, 0]]) * Utils.rotx(link_info[2, 0]) * Utils.trn(
        [link_info[3, 0], 0, 0]))
    Q02 = Q01 * (Utils.rotz(link_info[0, 0] + theta2) * Utils.trn([0, 0, link_info[1, 1]]) * Utils.rotx(link_info[2, 1]) * Utils.trn(
        [link_info[3, 1], 0, 0]))
    Q03 = Q02 * (Utils.rotz(link_info[0, 1] + theta3) * Utils.trn([0, 0, link_info[1, 2]]) * Utils.rotx(link_info[2, 2]) * Utils.trn(
        [link_info[3, 2], 0, 0]))
    Q04 = Q03 * (Utils.rotz(link_info[0, 2] + theta4) * Utils.trn([0, 0, link_info[1, 3]]) * Utils.rotx(link_info[2, 3]) * Utils.trn(
        [link_info[3, 3], 0, 0]))
    Q05 = Q04 * (Utils.rotz(link_info[0, 3] + theta5) * Utils.trn([0, 0, link_info[1, 4]]) * Utils.rotx(link_info[2, 4]) * Utils.trn(
        [link_info[3, 4], 0, 0]))
    Q06 = Q05 * (Utils.rotz(link_info[0, 4] + theta6) * Utils.trn([0, 0, link_info[1, 5]]) * Utils.rotx(link_info[2, 5]) * Utils.trn(
        [link_info[3, 5], 0, 0]))
    Q07 = Q06 * (Utils.rotz(link_info[0, 5] + theta7) * Utils.trn([0, 0, link_info[1, 6]]) * Utils.rotx(link_info[2, 6]) * Utils.trn(
        [link_info[3, 6], 0, 0]))
    Q08 = Q07 * (Utils.rotz(link_info[0, 6] + theta8) * Utils.trn([0, 0, link_info[1, 7]]) * Utils.rotx(link_info[2, 7]) * Utils.trn(
        [link_info[3, 7], 0, 0]))
    Q09 = Q08 * (Utils.rotz(link_info[0, 7]) * Utils.trn([0, 0, link_info[1, 8]]) * Utils.rotx(link_info[2, 8]) * Utils.trn(
        [link_info[3, 8], 0, 0]))

    link1_mth = Utils.inv_htm(Q01)
    link_3d_obj.append([
        # torre movel
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/58.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
        # cilindro conector
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/63.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
    ])

    link2_mth = Utils.inv_htm(Q02)
    link_3d_obj.append([
        # ligacao de cilindros longa
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/64.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh),
    ])

    link3_mth = Utils.inv_htm(Q03)
    link_3d_obj.append([
        # ligacao de cilindros longa
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/61.obj',
                scale=scale, htm=link3_mth, mesh_material=mesh),
    ])

    link4_mth = Utils.inv_htm(Q04)
    link_3d_obj.append([
        # cotovelo
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/14.obj',
                scale=scale, htm=link4_mth, mesh_material=mesh),
    ])

    link5_mth = Utils.inv_htm(Q05)
    link_3d_obj.append([
        # pá dobrada
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/36.obj',
                scale=scale, htm=link5_mth, mesh_material=mesh),
    ])

    link6_mth = Utils.inv_htm(Q06)
    link_3d_obj.append([
        # conecta pá com bumerangue
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/34.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh),
    ])

    link7_mth = Utils.inv_htm(Q07)
    link_3d_obj.append([
        # bumerangue
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/39.obj',
                scale=scale, htm=link7_mth, mesh_material=mesh),
    ])

    link8_mth = Utils.inv_htm(Q08)
    link_3d_obj.append([
        # envoltoria agulha
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/42.obj',
                scale=scale, htm=link8_mth, mesh_material=mesh),
    ])

    link9_mth = Utils.inv_htm(Q09)
    link_3d_obj.append([
        # guia
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/23.obj',
                    scale=scale, htm=link9_mth, mesh_material=mesh),
        # agulha
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/55.obj',
                scale=scale, htm=link9_mth, mesh_material=mesh),
        # abridor de garrafa
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/25.obj',
                scale=scale, htm=link9_mth, mesh_material=mesh),
        # botao
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/53.obj',
                scale=scale, htm=link9_mth, mesh_material=mesh),
    ])

    col_model = [[], [], [], [], [], [], [], [], []]

    col_model[0].append(Box(htm=Utils.trn([-0.02, 0.006*0, -0.43]) @ Utils.rotz(np.pi/2),
                            name=name + "_C0_0", width=0.14, height=0.4, depth=0.2, color="red", opacity=0.3))
    col_model[0].append(Cylinder(htm=Utils.trn([0, 0, -0.16]) @ Utils.rotz(3.14 / 2) @ Utils.rotx(3.14),
                                 name=name + "_C0_1", radius=0.075, height=0.19, color="red", opacity=0.3))

    col_model[1].append(Box(htm=Utils.trn([-0.200, 0, -0.11]) @ Utils.roty(np.pi/2),
                            name=name + "_C1_0", width=0.1, height=0.4, depth=0.12, color="green", opacity=0.3))
    col_model[1].append(Cylinder(htm=Utils.trn([0, 0, -0.12]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C1_1", radius=0.075, height=0.12, color="green", opacity=0.3))
    col_model[1].append(Cylinder(htm=Utils.trn([-0.415, 0, -0.11]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C1_2", radius=0.075, height=0.13, color="green", opacity=0.3))

    col_model[2].append(Box(htm=Utils.trn([-0.200, 0, 0.005]) @ Utils.roty(np.pi/2),
                            name=name + "_C2_0", width=0.1, height=0.4, depth=0.12, color="blue", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([0, 0, 0.01]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C2_1", radius=0.075, height=0.12, color="blue", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([-0.415, 0, 0.01]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C2_2", radius=0.075, height=0.13, color="blue", opacity=0.3))

    col_model[3].append(Cylinder(htm=Utils.trn([0, -0.025, 0.01]) @ Utils.rotx(-alpha4),
                                 name=name + "_C3_0", radius=0.075, height=0.12, color="magenta", opacity=0.3))
    col_model[3].append(Box(htm=Utils.trn([0, -0.02, 0.12]) @ Utils.roty(np.pi/2),
                            name=name + "_C3_1", width=0.2, height=0.14, depth=0.14, color="magenta", opacity=0.3))

    col_model[4].append(Box(htm=Utils.trn([0, -0.355, -0.05]) @ Utils.roty(np.pi/2),
                            name=name + "_C4_0", width=0.2, height=0.08, depth=0.05, color="orange", opacity=0.3))
    col_model[4].append(Box(htm=Utils.trn([0, -0.15, -0.125]) @ Utils.roty(np.pi/2),
                            name=name + "_C4_1", width=0.05, height=0.09, depth=0.4, color="orange", opacity=0.3))

    col_model[5].append(Box(htm=Utils.trn([-0.2, 0.025, -0.025]),
                            name=name + "_C5_0", width=0.25, height=0.14, depth=0.15, color="cyan", opacity=0.3))
    col_model[5].append(Box(htm=Utils.trn([-0.02, 0.025, -0.088]),
                            name=name + "_C5_1", width=0.14, height=0.02, depth=0.15, color="cyan", opacity=0.3))
    col_model[5].append(Box(htm=Utils.trn([-0.019, 0.025, 0.019]),
                            name=name + "_C5_2", width=0.14, height=0.02, depth=0.15, color="cyan", opacity=0.3))

    col_model[6].append(Box(htm=Utils.trn([-0.42, 0.08, -0.034]) @ Utils.rotz(np.pi/3),
                            name=name + "_C6_0", width=0.23, height=0.082, depth=0.13, color="#88264a", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.22, 0.08, -0.034]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C6_1", width=0.11, height=0.09, depth=0.39, color="#88264a", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.023, 0.01, 0.005]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C6_2", width=0.08, height=0.01, depth=0.03, color="#88264a", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.023, 0.01, -0.068]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C6_3", width=0.08, height=0.01, depth=0.03, color="#88264a", opacity=0.3))
    col_model[6].append(Cylinder(htm=Utils.trn([0, 0, -0.068]),
                                 name=name + "_C6_4", radius=0.035, height=0.01, color="#88264a", opacity=0.3))
    col_model[6].append(Cylinder(htm=Utils.trn([0, 0, 0.005]),
                                 name=name + "_C6_5", radius=0.035, height=0.01, color="#88264a", opacity=0.3))

    col_model[7].append(Box(htm=Utils.trn([-0.052, 0.0, -0.042]),
                            name=name + "_C7_0", width=0.02, height=0.31, depth=0.12, color="Brown", opacity=0.3))
    col_model[7].append(Cylinder(htm=Utils.trn([-0.1, -0.005, -0.001]) @ Utils.rotx(np.pi/2),
                                 name=name + "_C7_1", radius=0.035, height=0.056, color="Brown", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.08, -0.005, 0]),
                            name=name + "_C7_2", width=0.04, height=0.07, depth=0.056, color="Brown", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.03, -0.001, 0.073]),
                            name=name + "_C7_3", width=0.04, height=0.05, depth=0.04, color="Brown", opacity=0.3))
    col_model[7].append(Cylinder(htm=Utils.trn([0, -0.001, 0.074]),
                                 name=name + "_C7_4", radius=0.02, height=0.056, color="Brown", opacity=0.3))

    col_model[8].append(Box(htm=Utils.trn([-0.005, 0.0, -0.22]),
                            name=name + "_C8_0", width=0.07, height=0.4, depth=0.12, color="MidnightBlue", opacity=0.3))
    col_model[8].append(Cylinder(htm=Utils.trn([0, 0, 0.135]),
                                 name=name + "_C8_1", radius=0.01, height=0.4, color="MidnightBlue", opacity=0.3))

    links = []
    for i in range(n):
        links.append(Link(i, theta=link_info[0, i], d=link_info[1, i], alpha=link_info[2, i], a=link_info[3, i], joint_type=link_info[4, i],
                          list_model_3d=link_3d_obj[i]))
        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    #     1  2  3  4  5  6  7  8  9
    q0 = [0, 0, 0, 0, 0, 0, 0, 0, 0]
    htm_n_eef = Utils.trn([0, 0, 0.3323])  # np.identity(4)
    htm_base_0 = Utils.rotz(
        np.pi) * Utils.trn([-3.4*b1, b2, b3])  # np.identity(4)

    # Create joint limits
    joint_limits = np.matrix([[-0.1, 0.53], [-np.pi, np.pi], [-np.deg2rad(162), np.deg2rad(162)], [-np.pi, np.pi], [-np.pi, np.pi],
                             [-np.pi, np.pi], [-np.deg2rad(175), np.deg2rad(50)], [-np.deg2rad(108), np.deg2rad(70)], [-0.05, 0.079]])

    return links, base_3d_obj, htm_base_0, htm_n_eef, q0, joint_limits


def _create_davinci_arm4(color, opacity, name='davinci_arm4'):

    if not Utils.is_a_color(color):
        raise Exception(
            "The parameter 'color' should be a HTML-compatible color.")

    if (not Utils.is_a_number(opacity)) or opacity < 0 or opacity > 1:
        raise Exception(
            "The parameter 'opacity' should be a float between 0 and 1.")

    theta2 = np.deg2rad(164.8)
    theta3 = np.deg2rad(-131.8)
    theta4 = np.deg2rad(-33)
    theta5 = np.deg2rad(-91)
    theta6 = np.deg2rad(94.5)
    theta7 = np.deg2rad(-169.82)
    theta8 = np.deg2rad(37.1)

    d3 = 96e-3 * 1.25
    d4 = 96e-3 * 3.68
    d5 = 0.919
    d8 = -3.4e-2
    d9 = 0.069

    alpha4 = -np.deg2rad(90 + 63.1)
    alpha5 = -np.pi/2
    alpha8 = np.deg2rad(90.7)

    a2 = 0.3
    a3 = 0.3
    a5 = -8e-4
    a6 = 0.27
    a7 = 0.473
    a8 = 0.115

    # Passive joints 1, 2, 3, 4, 7, 8
    # Active joints 5, 6, 9
    link_info = np.array([
        # "theta" rotation in z
        [0,  0,   0,       0,      0,      0,      0,       0, 0],
        # 0,  "d" translation in z
        [0,  0,  d3,      d4,     d5,      0,      0,      d8, d9],
        # 0,  "alfa" rotation in x
        [0,  0,   0,  alpha4, alpha5,      0,      0,  alpha8, 0],
        # 0,  "a" translation in x
        [0, a2,  a3,       0,     a5,     a6,     a7,      a8, 0],
        # 0,  joint type
        [1,  0,   0,       0,      0,      0,      0,       0, 1]
    ])

    scale = 1
    n = link_info.shape[1]
    base_3d_obj = []
    link_3d_obj = []
    mesh = MeshMaterial(metalness=0.5, clearcoat=0, roughness=0.5, normal_scale=[
                        0.5, 0.5], color=color, opacity=opacity, side="DoubleSide")
    b1 = 0.099
    b2 = 0.0922
    b3 = 1.54

    Q00 = Utils.trn([b1, b3, -b2]) * Utils.rotx(-np.pi/2) * \
        Utils.rotz(np.pi)  # change reference frame
    Q01 = Q00 * (Utils.rotz(link_info[0, 0]) * Utils.trn([0, 0, link_info[1, 0]]) * Utils.rotx(link_info[2, 0]) * Utils.trn(
        [link_info[3, 0], 0, 0]))
    Q02 = Q01 * (Utils.rotz(link_info[0, 0] + theta2) * Utils.trn([0, 0, link_info[1, 1]]) * Utils.rotx(link_info[2, 1]) * Utils.trn(
        [link_info[3, 1], 0, 0]))
    Q03 = Q02 * (Utils.rotz(link_info[0, 1] + theta3) * Utils.trn([0, 0, link_info[1, 2]]) * Utils.rotx(link_info[2, 2]) * Utils.trn(
        [link_info[3, 2], 0, 0]))
    Q04 = Q03 * (Utils.rotz(link_info[0, 2] + theta4) * Utils.trn([0, 0, link_info[1, 3]]) * Utils.rotx(link_info[2, 3]) * Utils.trn(
        [link_info[3, 3], 0, 0]))
    Q05 = Q04 * (Utils.rotz(link_info[0, 3] + theta5) * Utils.trn([0, 0, link_info[1, 4]]) * Utils.rotx(link_info[2, 4]) * Utils.trn(
        [link_info[3, 4], 0, 0]))
    Q06 = Q05 * (Utils.rotz(link_info[0, 4] + theta6) * Utils.trn([0, 0, link_info[1, 5]]) * Utils.rotx(link_info[2, 5]) * Utils.trn(
        [link_info[3, 5], 0, 0]))
    Q07 = Q06 * (Utils.rotz(link_info[0, 5] + theta7) * Utils.trn([0, 0, link_info[1, 6]]) * Utils.rotx(link_info[2, 6]) * Utils.trn(
        [link_info[3, 6], 0, 0]))
    Q08 = Q07 * (Utils.rotz(link_info[0, 6] + theta8) * Utils.trn([0, 0, link_info[1, 7]]) * Utils.rotx(link_info[2, 7]) * Utils.trn(
        [link_info[3, 7], 0, 0]))
    Q09 = Q08 * (Utils.rotz(link_info[0, 7]) * Utils.trn([0, 0, link_info[1, 8]]) * Utils.rotx(link_info[2, 8]) * Utils.trn(
        [link_info[3, 8], 0, 0]))

    link1_mth = Utils.inv_htm(Q01)
    link_3d_obj.append([
        # torre movel
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/57.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
        # cilindro conector
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/60.obj',
                scale=scale, htm=link1_mth, mesh_material=mesh),
    ])

    link2_mth = Utils.inv_htm(Q02)
    link_3d_obj.append([
        # ligacao de cilindros curta
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/15.obj',
                scale=scale, htm=link2_mth, mesh_material=mesh),
    ])

    link3_mth = Utils.inv_htm(Q03)
    link_3d_obj.append([
        # ligacao de cilindros curta
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/66.obj',
                scale=scale, htm=link3_mth, mesh_material=mesh),
    ])

    link4_mth = Utils.inv_htm(Q04)
    link_3d_obj.append([
        # cotovelo
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/16.obj',
                scale=scale, htm=link4_mth, mesh_material=mesh),
    ])

    link5_mth = Utils.inv_htm(Q05)
    link_3d_obj.append([
        # pá dobrada
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/37.obj',
                scale=scale, htm=link5_mth, mesh_material=mesh),
    ])

    link6_mth = Utils.inv_htm(Q06)
    link_3d_obj.append([
        # conecta pá com bumerangue
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/33.obj',
                scale=scale, htm=link6_mth, mesh_material=mesh),
    ])

    link7_mth = Utils.inv_htm(Q07)
    link_3d_obj.append([
        # bumerangue
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/41.obj',
                scale=scale, htm=link7_mth, mesh_material=mesh),
    ])

    link8_mth = Utils.inv_htm(Q08)
    link_3d_obj.append([
        # envoltoria agulha
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/28.obj',
                scale=scale, htm=link8_mth, mesh_material=mesh),
    ])

    link9_mth = Utils.inv_htm(Q09)
    link_3d_obj.append([
        # guia
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/29.obj',
                    scale=scale, htm=link9_mth, mesh_material=mesh),
        # agulha
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/31.obj',
                scale=scale, htm=link9_mth, mesh_material=mesh),
        # base agulha
        Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/30.obj',
                scale=scale, htm=link9_mth, mesh_material=mesh),
        # cabo longo
        # Model3D(url='https://raw.githubusercontent.com/viniciusmgn/uaibot_content/master/contents/DaVinci/32.obj',
        #        scale=scale, htm=link9_mth, mesh_material=mesh),
    ])

    col_model = [[], [], [], [], [], [], [], [], []]

    col_model[0].append(Box(htm=Utils.trn([0, -0.03, -0.33]),
                            name=name + "_C0_0", width=0.14, height=0.4, depth=0.2, color="red", opacity=0.3))
    col_model[0].append(Cylinder(htm=Utils.trn([0, 0, -0.154]) @ Utils.rotz(3.14 / 2) @ Utils.rotx(3.14),
                                 name=name + "_C0_1", radius=0.075, height=0.19, color="red", opacity=0.3))

    col_model[1].append(Box(htm=Utils.trn([-0.15, 0, 0]) @ Utils.roty(np.pi/2),
                            name=name + "_C1_0", width=0.11, height=0.3, depth=0.12, color="green", opacity=0.3))
    col_model[1].append(Cylinder(htm=Utils.trn([0, 0, -0.001]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C1_1", radius=0.075, height=0.123, color="green", opacity=0.3))
    col_model[1].append(Cylinder(htm=Utils.trn([-0.3, 0, 0.002]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C1_2", radius=0.075, height=0.12, color="green", opacity=0.3))

    col_model[2].append(Box(htm=Utils.trn([-0.15, 0, 0.0]) @ Utils.roty(np.pi/2),
                            name=name + "_C2_0", width=0.11, height=0.3, depth=0.12, color="blue", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([0, 0, -0.025]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C2_1", radius=0.075, height=0.08, color="blue", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([-0.3, 0, 0.002]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C2_2", radius=0.075, height=0.12, color="blue", opacity=0.3))
    col_model[2].append(Cylinder(htm=Utils.trn([0, 0, 0.033]) @ Utils.rotz(np.pi/2) @ Utils.rotx(3.14),
                                 name=name + "_C2_3", radius=0.062, height=0.05, color="blue", opacity=0.3))

    col_model[3].append(Cylinder(htm=Utils.rotx(-alpha4) @ Utils.trn([0, 0, 0.101-d4]),
                                 name=name + "_C3_0", radius=0.075, height=0.08, color="magenta", opacity=0.3))
    col_model[3].append(Box(htm=Utils.trn([0, 0.02, 0.42]) @ Utils.roty(np.pi/2),
                            name=name + "_C3_1", width=0.23, height=0.14, depth=0.12, color="magenta", opacity=0.3))
    col_model[3].append(Box(htm=Utils.rotx(-alpha4) @ Utils.trn([0, 0.072, -0.259]),
                            name=name + "_C3_2", width=0.14, height=0.07, depth=0.2, color="magenta", opacity=0.3))

    col_model[4].append(Box(htm=Utils.trn([0, 0.355, -0.05]) @ Utils.roty(np.pi/2),
                            name=name + "_C4_0", width=0.2, height=0.08, depth=0.05, color="orange", opacity=0.3))
    col_model[4].append(Box(htm=Utils.trn([0, 0.15, -0.1288]) @ Utils.roty(np.pi/2),
                            name=name + "_C4_1", width=0.05, height=0.09, depth=0.4, color="orange", opacity=0.3))

    col_model[5].append(Box(htm=Utils.trn([-0.2, 0.025, -0.03]),
                            name=name + "_C5_0", width=0.25, height=0.14, depth=0.15, color="cyan", opacity=0.3))
    col_model[5].append(Box(htm=Utils.trn([-0.018, 0.02, -0.09]),
                            name=name + "_C5_1", width=0.14, height=0.02, depth=0.15, color="cyan", opacity=0.3))
    col_model[5].append(Box(htm=Utils.trn([-0.019, 0.02, 0.015]),
                            name=name + "_C5_2", width=0.14, height=0.02, depth=0.15, color="cyan", opacity=0.3))

    col_model[6].append(Box(htm=Utils.trn([-0.42, 0.08, -0.0372]) @ Utils.rotz(np.pi/3),
                            name=name + "_C6_0", width=0.23, height=0.082, depth=0.13, color="#88264a", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.22, 0.08, -0.034]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C6_1", width=0.11, height=0.09, depth=0.39, color="#88264a", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.023, 0.008, 0.002]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C6_2", width=0.08, height=0.01, depth=0.03, color="#88264a", opacity=0.3))
    col_model[6].append(Box(htm=Utils.trn([-0.023, 0.008, -0.071]) @ Utils.rotz(np.pi*7/18),
                            name=name + "_C6_3", width=0.08, height=0.01, depth=0.03, color="#88264a", opacity=0.3))
    col_model[6].append(Cylinder(htm=Utils.trn([0, 0.001, -0.071]),
                                 name=name + "_C6_4", radius=0.035, height=0.01, color="#88264a", opacity=0.3))
    col_model[6].append(Cylinder(htm=Utils.trn([0, 0, 0.002]),
                                 name=name + "_C6_5", radius=0.035, height=0.01, color="#88264a", opacity=0.3))

    col_model[7].append(Box(htm=Utils.trn([-0.065, 0, -0.18]),
                            name=name + "_C7_0", width=0.028, height=0.63, depth=0.12, color="Brown", opacity=0.3))
    col_model[7].append(Cylinder(htm=Utils.trn([-0.115, 0.001, -0.0013]) @ Utils.rotx(np.pi/2),
                                 name=name + "_C7_1", radius=0.035, height=0.056, color="Brown", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.09, -0.005*0, 0.0]),
                            name=name + "_C7_2", width=0.04, height=0.07, depth=0.056, color="Brown", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.031, -0.001*0, 0.095]),
                            name=name + "_C7_3", width=0.048, height=0.09, depth=0.04, color="Brown", opacity=0.3))
    col_model[7].append(Cylinder(htm=Utils.trn([0.01, 0.001*0, 0.112]),
                                 name=name + "_C7_4", radius=0.025, height=0.045, color="Brown", opacity=0.3))
    col_model[7].append(Box(htm=Utils.trn([-0.1, 0, -0.306]),
                            name=name + "_C7_5", width=0.04, height=0.378, depth=0.12, color="Brown", opacity=0.3))

    col_model[8].append(Box(htm=Utils.trn([-0.002, 0.0, -0.52]),
                            name=name + "_C8_0", width=0.09, height=0.36, depth=0.12, color="MidnightBlue", opacity=0.3))
    col_model[8].append(Cylinder(htm=Utils.trn([0, 0, -0.109]),
                                 name=name + "_C8_1", radius=0.013, height=0.6, color="MidnightBlue", opacity=0.3))

    links = []
    for i in range(n):
        links.append(Link(i, theta=link_info[0, i], d=link_info[1, i], alpha=link_info[2, i], a=link_info[3, i], joint_type=link_info[4, i],
                          list_model_3d=link_3d_obj[i]))
        for j in range(len(col_model[i])):
            links[i].attach_col_object(col_model[i][j], col_model[i][j].htm)

    # Define initial configuration
    #     1  2  3  4  5  6  7  8  9
    q0 = [0, 0, 0, 0, 0, 0, 0, 0, 0]
    htm_n_eef = Utils.trn([0, 0, 0.1899 - d9])  # np.identity(4)
    htm_base_0 = Utils.trn([-b1, -b2, b3])  # np.identity(4)

    # Create joint limits
    joint_limits = np.matrix([[-0.423, 0.21], [-np.pi, np.pi], [-np.deg2rad(158), np.deg2rad(156)], [-np.deg2rad(200), np.deg2rad(
        20)], [-np.pi, np.pi], [-np.pi, np.pi], [-np.deg2rad(170), np.deg2rad(50)], [-np.deg2rad(108), np.deg2rad(60)], [0, 0.295]])

    return links, base_3d_obj, htm_base_0, htm_n_eef, q0, joint_limits